#ifndef BIHCLASS
#define BIHCLASS

#include "Primitive.h"
#include "RayAccel.h"

class Ray; // forward declaration of Ray class
class BBox;

#define EPSILON_OS  0.03    // (BIH Size)
#define EPSILON_OI  0.00005 // (BIH Include)
//#define SHOWTRAV 1


class BIH : public RayAccel {
private:

  // cell
  const static int BIH_X=0;
  const static int BIH_Y=1;
  const static int BIH_Z=2;
  const static int BIH_LEAF=3;
  
  typedef struct node_st {
    int index;         // to child nodes  (inner nodes)
                       // to first object (leaf nodes)
    // FIXME: next line can be included in lower bits of index
    int nodeType;
    union {
      float clip[2];   // clipping planes (inner nodes)
      int items;       // number of items (leaf nodes)
    };
    
    bool isLeaf() const { return nodeType == BIH_LEAF; }
    bool isSplit() const { return nodeType != BIH_LEAF; }
    int splitAxis() const { return nodeType; }
    int setAxis(int a) { nodeType = a; }
    void setLeaf() { nodeType = BIH_LEAF; }
  } node;
  
  // parameters struct
  typedef struct treeParam {
    Ray *r_p;                 // ray
    SbVec3f invDir;
    double t0, t1;
  } treeParam;
  
  int bihDebug;
  int maxItems;
  int maxDepth;
  node *root;
  PrimList primList;
  vector<node> nodeList;
  unsigned int rayId;
  pthread_mutex_t lock;

  // temporary!
  int rays;

  // also keep bounding sphere of specular primitives
  SbVec3f shiny_min, shiny_max;
  
  void printStatsTrav(int nodeIdx, int &numInner,
                      int &numLeafs, int &maxPrims,
                      int &minPrims, int &totPrims);
  void refineNode();

public:

  BIH();
  ~BIH();

  // Primitive methods
  PrimType getType() const { return PRIMTYPE_BIH; }
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  SbVec3f getMinShiny() const { return shiny_min; }
  SbVec3f getMaxShiny() const { return shiny_max; }
  
  void reset();

  int add(Primitive *t){

    // add primitives
    numPrims++;
    primList.add(t);    
  }
  void reserve(int n) { primList.reserve(n); }
  void resizeToFit() { primList.resizeToFit(); }

  Material* getMat(const Ray &r) {
    return primList.size() ? primList[0]->getMat(r) : Primitive::getMat(r);
  }
  const Material* getMat(const Ray &r) const {
    return primList.size() ? primList[0]->getMat(r) : Primitive::getMat(r);
  }

  void printStats();
  void setParams(int maxD, int maxI, int debug);
  void refine(int simpleDiv = 0);
};

#endif /* BIHCLASS */
