//WWWWW'''''''''''WW'''''''''''''''''WW''''WW''''''''''
//WW                                 WWW  WWW          
//WWWW  WWW  WW W WW W WW  WW   WWW  W WWW WW  WWW WWW 
//WW    WW W WW W WW WW   WW W  WW W W  W  WW W WW WW W
//WW    WW W WW W WW WW   WW W  WW W W     WW W WW WW W
//WWWWW WW W  WW  WW WW    WW   WW W W     WW  WWW WWW 
//                                                 WW  
//'''''''''''''''''''''''''''''''''''''''''''''''''WW''

#ifndef ENVIRONMAPCLASS
#define ENVIRONMAPCLASS

#include <string>
#include <Inventor/SbColor.h>

using namespace std;

class GlobalVars;
class Ray;

class EnvironMap {

  public:
  typedef enum { TEXTURE, SOLID, SKY } envType_t;
  
  protected:
  
  GlobalVars *gv;

  private:
  
  SbColor *data;
  SbColor *data3;  // downsampled lat/long graymap for debugging
  int dim[2];
  int dim2[2];
  int dimll[2];
  float **pweights, *cweights;  // pixel weights (columns), column weights
  double *samples1, *samples2;
  int samples;
  float *sint; // precomputed sin(theta)
  envType_t envType;          // the type of environment map
  bool isSingleColor;
  
  public:
  
  SbColor Lavg;               // the average incoming radiance (RGB)
  float Lavgavg;              // average of Lavg (float)
  bool valid;                 // state of environment map
  SbMatrix worldToTextureRot; // rotation to turn world dir into texture-space dir
  SbMatrix textureToWorldRot; // rotation to turn texture-space dir into world dir
  
  EnvironMap();
  virtual ~EnvironMap();
  
  virtual void init(GlobalVars *gv_) { init(gv_, 0x0, 0, 0, 0x0); }
  virtual void init(GlobalVars *gv_, float *inputPixels,
                    int pixX, int pixY, SbRotation *rot);
  virtual SbColor impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple=0);
  virtual SbColor volImpSample(Ray &r, int simple);  
  virtual float getSample(double eps1, double eps2,
		  SbVec3f &dir, SbColor &c);
  virtual SbColor getRadiance(SbVec3f &dir);
  virtual SbColor getEmissiveFlux(float receiverArea);
  bool setSingleColor(bool b){ isSingleColor = b; }

  float getPdf(const SbVec3f &l);

  envType_t &getType() { return envType; }
  const envType_t &getType() const { return envType; }
  void setRotateSky(const SbRotation &rot);
  void writeDataHdr(const string &file);
  void writeDataExr(const string &file);
  void writeData3();
};

#endif /* ENVIRONMAPCLASS */
  
