#include "FontWriter.h"
#include <Inventor/SbColor.h>

// source stolen from FreeType example1.c
// http://freetype.sourceforge.net/freetype2/docs/tutorial/example1.c

#include <stdio.h>
#include <string.h>
#include <string>
#include <math.h>
#include <iostream>

#include <ft2build.h>
#include FT_FREETYPE_H

using namespace std;

void
draw_bitmap( FT_Bitmap*  bitmap,
             FT_Int      x,
             FT_Int      y,
             const int imgDim[2], SbColor *pix, SbColor color)
{
  FT_Int  i, j, p, q;
  FT_Int  x_max = x + bitmap->width;
  FT_Int  y_max = y + bitmap->rows;

  
  for ( i = x, p = 0; i < x_max; i++, p++ )
  {
    for ( j = y, q = 0; j < y_max; j++, q++ )
    {
      if ( i<0 || i >= imgDim[0] ||
           j<0 || j >= imgDim[1] )
        continue;

      float val = (bitmap->buffer[q * bitmap->width + p]) * (1/255.);

      // composite pixel
      int idx = (imgDim[1]-j-1)*imgDim[0]+i;
      pix[idx] = pix[idx]*(1-val) + color*val;
        
    }
  }
}

void FontWriter::write(const int imgDim[2], SbColor *pix, const string &str) {

  FT_Library    library;
  FT_Face       face;

  FT_GlyphSlot  slot;
  FT_UInt       glyph_index;
  FT_Vector     pen;                    /* untransformed origin  */
  FT_Error      error;

  const char*         filename;
  const char*         filename2;
  const char*         text;

  double        angle;
  int           target_height;
  int           n, num_chars;


  text          = (const char *)str.c_str();
  num_chars     = str.length();
  target_height = imgDim[1];

  error = FT_Init_FreeType( &library );              /* initialize library */
  /* error handling omitted */

  //filename = "/usr/share/fonts/truetype/freefont/FreeMono.ttf";
  filename = "/usr/share/fonts/truetype/freefont/FreeSansOblique.ttf";
  filename2 = "/usr/share/X11/fonts/TTF/luxirb.ttf";
  error = FT_New_Face( library, filename, 0, &face ); /* create face object */

  if (error){
    fprintf(stderr, "Watermark: Error opening font: %s. Trying another font.\n",
	    filename);
    filename = filename2;
    error = FT_New_Face( library, filename, 0, &face ); /* create face object */
  }
  if (error) {
    fprintf(stderr, "Watermark: Error opening font: %s. "
	    "Aborting watermark.\n", filename);
    return;
  }

  /* use Xpt at 100dpi */
  int pt = 7;
  error = FT_Set_Char_Size( face, pt * 64, 0,
                            100, 0 );                /* set character size */
  /* error handling omitted */

  slot = face->glyph;

  /* the pen position in 26.6 cartesian space coordinates; */
  /* start at (300,200) relative to the upper left corner  */
  pen.x = imgDim[0]-2;
  pen.y = imgDim[1]-1;
  
  // compute size
  for ( n = 0; n < num_chars; n++ )
  {
    
    /* load glyph image into the slot (erase previous one) */
    error = FT_Load_Char( face, text[n], FT_LOAD_RENDER );
    if ( error )
      continue;                 /* ignore errors */

    pen.x -= slot->advance.x >> 6;
  }

  FT_Vector origPen = pen;

  // draw dark bg
  pen.x++;
  pen.y++;
  for ( n = 0; n < num_chars; n++ )
  {
    
    /* load glyph image into the slot (erase previous one) */
    error = FT_Load_Char( face, text[n], FT_LOAD_RENDER );
    if ( error )
      continue;                 /* ignore errors */

    /* now, draw to our target surface (convert position) */
    draw_bitmap( &slot->bitmap,
                 pen.x + slot->bitmap_left,
                 pen.y - slot->bitmap_top,
                 imgDim, pix, SbColor(0.25, 0.25, 0.25));

    /* increment pen position */
    pen.x += slot->advance.x >> 6;
  }
  
  // draw white fg
  pen = origPen;
  for ( n = 0; n < num_chars; n++ )
  {
    
    /* load glyph image into the slot (erase previous one) */
    error = FT_Load_Char( face, text[n], FT_LOAD_RENDER );
    if ( error )
      continue;                 /* ignore errors */

    /* now, draw to our target surface (convert position) */
    draw_bitmap( &slot->bitmap,
                 pen.x + slot->bitmap_left,
                 pen.y - slot->bitmap_top,
                 imgDim, pix, SbColor(1,1,1));

    /* increment pen position */
    pen.x += slot->advance.x >> 6;
  }
  
  FT_Done_Face    ( face );
  FT_Done_FreeType( library );

  return;
}
