#ifndef FRACTAL_H
#define FRACTAL_H

#include "Primitive.h"

class SoLinearFractal;

class Fractal : public Primitive {

private:
  Primitive *prim;
  vector<SbMatrix> xforms;
  vector<float> scales;
  int maxDepth;
  SbVec3f bsPos;  // bounding sphere
  float bsRad;   
  bool onlyIntersectBV;
  bool skipIthXform;
  bool intersectIntermediate;

public:
  Fractal(Primitive *p, SoLinearFractal *soFract=NULL);
  void init(SoLinearFractal *soFract=NULL);

  ~Fractal(){}

  void deletePrim(){ delete prim; }

  virtual PrimType getType() const { return PRIMTYPE_FRACTAL; }
  virtual string getTypeName() const { return "Fractal"; }

  void loadXforms(const char *file);
  bool intersectSphere(Ray &r, const SbVec3f &pos, const float rad,
                       const double &t0, const double &t1,
                       double &tNear, double &tFar);

  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  
  virtual SbVec3f  getMin() const {
    return bsPos-SbVec3f(bsRad, bsRad, bsRad);
  }
  virtual SbVec3f  getMax() const {
    return bsPos+SbVec3f(bsRad, bsRad, bsRad);
  }

  virtual Material* getMat() { return prim->getMat(); }
  virtual const Material* getMat() const { return prim->getMat(); }
  virtual Material* getMat(const Ray &r) { return prim->getMat(r); }
  virtual const Material* getMat(const Ray &r) const { return prim->getMat(r); }
  
  virtual SbVec3f getNormal(      const SbVec3f &bc) const{
    return bc;
  }
  
  virtual SbColor getIntDiffColor(const Ray &r) const { return prim->getIntDiffColor(r);}
  virtual SbColor getIntEmission(const Ray &r) const { return prim->getIntEmission(r);}
  
//   virtual bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer) const{
//     if (!prim->getDerivs(bc, pDer)) return false;
//     SbMatrix tr = mi.transpose();
//     tr.multDirMatrix(pDer.dPdU, pDer.dPdU);
//     tr.multDirMatrix(pDer.dPdV, pDer.dPdV);
//     tr.multDirMatrix(pDer.dNdU, pDer.dNdU);
//     tr.multDirMatrix(pDer.dNdV, pDer.dNdV);
//     return true;
//   }
  
};


#endif


#ifndef SOLINEARFRACTALCLASS
#define SOLINEARFRACTALCLASS

#include <iostream>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/nodes/SoSubNode.h>
#include <Inventor/fields/SoSFColor.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFVec2f.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFBool.h>

using namespace std;

//   SoLinearFractal {
//     fields [ SFInt32 depth, SFString xformsFile,
//              SFBool isectBVOnly, SFBool isectIntermediate,
//              SFBool skipIthXform ]
//     depth 14
//     xformsFile "test.iv"
//     skipIthXform TRUE
//   }


/*

  Another possibility:

Fractal {
  transforms "ompf_xforms3.iv"
  prob [ .7 .3 ... ]
  Separator {
    (model)
  }
  Separator {
    (model)
  }
  ...
}
  
 */

class SoLinearFractal : public SoNode {

  SO_NODE_HEADER(SoLinearFractal);

  public:

  SoSFInt32 depth;
  SoSFString xformsFile;
  SoSFBool isectBVOnly;
  SoSFBool isectIntermediate;
  SoSFBool skipIthXform;
  
  static void initClass();
  friend ostream& operator<< (ostream& os, const SoMatPlus& m);

  SoLinearFractal();

  protected:

  virtual void GLRender(SoGLRenderAction *action);
  virtual void callback(SoCallbackAction *action);
  virtual void doAction(SoAction *action, const char *s);

  private:

  virtual ~SoLinearFractal();
};

#endif /* SOLINEARFRACTALCLASS */
