//WW''W''''''WWW'''WW''''''''''''
//WW  W       WW   WW            
//WW  W  WWW  WW  WWWW  WW   WWW 
//WWWWW W WW  WW   WW  WW W  WW W
//WW  W W WW  WW   WW  WW W  WW W
//WW  W  WWW WWWW  WW   WW   WW W
//                               
//'''''''''''''''''''''''''''''''

#include <stdlib.h>
#include <iostream>
#include <time.h>
#include <math.h>
#include "Halton.h"

using namespace std;


// Hammersley sequence
//
// i/N, b_0=2, b_1=3, b_2=5, b_3=7, b_4=11, b_5=13, ...
//
// b_d, d = dimension, and index of prime

vector<int> Halton::primes;
vector<vector<unsigned short> > Halton::sigmas;

void Halton::Number(long i, int base){
  double f = inv_base = 1.0/base;
  value = 0.0;
  while (i>0){
    value += f* (double)(i % base);
    i /= base;
    f *= inv_base;
  }
}

// I made this... sigma is permuted set of 0...base-1
// for digit scrambling
void Halton::Number(long i, int base, unsigned short *sigma){
  double f = inv_base = 1.0/base;
  value = 0.0;
  while (i>0){
    value += f* (double)sigma[i % base];
    i /= base;
    f *= inv_base;
  }
}

void Halton::Next(){
  double r = 1.0 - value - 0.0000000001;
  if (inv_base < r) value += inv_base;
  else {
    double h = inv_base, hh;
    do {
      hh = h;
      h *= inv_base;
    } while (h >= r);
    value += hh + h - 1.0;
  }
}

namespace {

  vector<unsigned short> sigmaB(int b){

    if (b<2) {
      cerr << __FILE__ << ":L" << __LINE__ << ": Serious error!\n";
    }

    if (b==2){
      vector<unsigned short> ret(2);
      ret[0] = 0;
      ret[1] = 1;
      return ret;
    } else if (b%2 == 0) {  // b even
      vector<unsigned short> prev = sigmaB(b/2);
      int prevSize = prev.size();
      vector<unsigned short> ret(prevSize*2);
      for (int i=0; i<prevSize; i++){
        ret[i] = 2*prev[i];
        ret[prevSize+i] = 2*prev[i]+1;
      }
      return ret;
    } else {
      vector<unsigned short> ret = sigmaB(b-1);
      int mid = (b-1)/2;
      for (int i=0; i<ret.size(); i++)
        if (ret[i]>=mid)
          ret[i]++;
      ret.reserve(ret.size()+1);
      ret.insert(vector<unsigned short>::iterator(ret.begin()+ret.size()/2), mid);
      return ret;
    }
  }
}


void Halton::initClass() {

  primes.resize(numPrimes);
  sigmas.reserve(numPrimes);

  // compute primes using sieve
  int numCands = numPrimes*50; // Conservative. 1000th prime is 1000*8 (7919)
  int sqrtCands = (int)sqrt(numCands)+1;
  vector<bool> cands(numCands, true);
  cands[0]=cands[1]=false;
  for (int i=2; i<sqrtCands; i++){
    if (!cands[i]) continue;
    for (int j=i*2; j<numCands; j+=i)
      cands[j] = false;
  }

  // store first n primes
  int j=2;
  for (int i=0; i<numPrimes; i++){

    // find next prime
    while (!cands[j])
      j++;

    primes[i] = j++;
  }

  // compute sigmas
  // from Strictly Deterministic Sampling page 5
  for (int i=0; i<numPrimes; i++){
    vector<unsigned short> s = sigmaB(primes[i]);
    sigmas.push_back(s);
  }
}




//WW''''WW''''''WWW'''WW''WW'WW''W''''''WWW'''WW''''''''''''
//WWW  WWW       WW   WW     WW  W       WW   WW            
//W WWW WW W WW  WW  WWWW WW WW  W  WWW  WW  WWWW  WW   WWW 
//W  W  WW W WW  WW   WW  WW WWWWW W WW  WW   WW  WW W  WW W
//W     WW W WW  WW   WW  WW WW  W W WW  WW   WW  WW W  WW W
//W     WW  WWW WWWW  WW  WW WW  W  WWW WWWW  WW   WW   WW W
//                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

MultiHalton::MultiHalton(){

  // initialize
  // srand48(time(NULL));
  for (int i=0; i<dim; i++)
    h[i].Number(i+3/*lrand48()*/, prime(i));

  // precompute table
  for (int i=0; i<dim; i++){   // loop over series

    // start new series
    preh[i].Number(i+3, prime(i));

    // fill table
    for (int j=0; j<predim; j++){  // loop over entry
      preh[i].Next();
      table[i][j] = preh[i];
    }
  }
}

double MultiHalton::next(int d){
  h[d].Next();
  return(h[d]);
}

double MultiHalton::prenext(int d){
  int i = (int)(drand48()*predim);
  return(table[d][i]);
}

const int MultiHalton::primes[MultiHalton::numPrimes] = { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229 };


//WW''W''''''WWW''WWW'''''''''''
//WW W        WW   WW           
//WWW    WW   WW   WW   WW  W WW
//WW W  WWWW  WW   WW  WWWW WW  
//WW  W WW    WW   WW  WW   WW  
//WW  W  WWW WWWW WWWW  WWW WW  
//                              
//''''''''''''''''''''''''''''''


//Quasi-Monte Carlo Integration (1995)  (Make Corrections)  (18 citations)
//William J. Morokoff, Russel E. Caflisch
// says Sobol is better for dim>6

// From Efficient Multidimensional Sampling
// by Thomas Kollig and Alexander Keller

double RI_vdC(uint bits, uint r){
  bits = ( bits << 16)
    | (bits >> 16);
  bits = ((bits & 0x00ff00ff) << 8)
    | ((bits & 0xff00ff00) >> 8);
  bits = ((bits & 0x0f0f0f0f) << 4)
    | ((bits & 0xf0f0f0f0) >> 4);
  bits = ((bits & 0x33333333) << 2)
    | ((bits & 0xcccccccc) >> 2);
  bits = ((bits & 0x55555555) << 1)
    | ((bits & 0xaaaaaaaa) >> 1);

  bits ^= r;

  return (double) bits / (double) 0x100000000LL;
}

double RI_S(uint i, uint r){
  for (uint v = 1U<<31; i; i >>= 1, v ^= v>>1)
    if (i & 1)
      r ^= v;

  return (double) r / (double) 0x100000000LL;
}

double RI_LP(uint i, uint r){
   for (uint v = 1U<<31; i; i >>= 1, v |= v>>1)
     if (i & 1)
       r ^= v;
   
   return (double) r / (double) 0x100000000LL;
}
