#include "Isosurface.h"
#include "Main.h"
#include "debug.h"

Material Isosurface::defaultMat;

Isosurface::Isosurface(float vertFOV, int vertRes) {
  c = 0;
  int lgVal = 100;//1.5;
  min = SbVec3f(-lgVal, -lgVal, -lgVal);
  max = SbVec3f(lgVal, lgVal, lgVal);
  mat = &defaultMat;
  errorAt1 = tanf(vertFOV*M_PI/180. / vertRes);

//   noise.octaves = 10;
//   noise.amplitude = 1;
//   float scale = 3;
//   noise.frequency = SbVec3f(scale,scale,scale);
//   //noise.translation = -offset;
//   //noise.bias = clamp(bias,0,1);
//   //noise.gain = clamp(gain,0,1);
//   noise.ampJump = 1;
//   noise.freqJump = 2;
//   noise.sign=false;

//   noise.octaves = 8;
//   noise.amplitude = .99/1.92651;
//   float scale = 2.638*.173;
//   noise.frequency = SbVec3f(scale,scale,scale);
//   noise.bias = .4257;
//   noise.gain = .8769;
//   noise.translation = -SbVec3f(.3742, .135, -.5825);
//   noise.freqJump = 2.029;
//   noise.ampJump = 1;
//   noise.sign=true;

   noise.octaves = 1;
//   noise.amplitude = .99/1.92651;
//   float scale = 2.638*.173;
//   noise.frequency = SbVec3f(scale,scale,scale);
//   noise.bias = .4257;
//   noise.gain = .8769;
//   noise.translation = -SbVec3f(.3742, .135, -.5825);
//   noise.freqJump = 2.029;
//   noise.ampJump = 1;

  noise.octaves = 3;
  noise.funcPtr = &Noise::fBm;
   
  noise.noise(SbVec3f(0,0,0)); // initialize exponent array
}

void Isosurface::clipToVolume(const Ray &r, double &t0, double &t1){

  // intersect line with box
  SbVec3f tmin, tmax;
  float eps = 1e-25;  // to prevent divide by zero
  tmin[0] = (min[0]-r.pos[0])/(r.dir[0]+eps); // x=x0+rt  ->  t=(x-x0)/r
  tmax[0] = (max[0]-r.pos[0])/(r.dir[0]+eps); 
  tmin[1] = (min[1]-r.pos[1])/(r.dir[1]+eps); 
  tmax[1] = (max[1]-r.pos[1])/(r.dir[1]+eps); 
  tmin[2] = (min[2]-r.pos[2])/(r.dir[2]+eps); 
  tmax[2] = (max[2]-r.pos[2])/(r.dir[2]+eps);

  // swap min/max
  float tmp;
  if (r.dir[0]<0.0){ tmp = tmin[0]; tmin[0] = tmax[0]; tmax[0] = tmp; }
  if (r.dir[1]<0.0){ tmp = tmin[1]; tmin[1] = tmax[1]; tmax[1] = tmp; }
  if (r.dir[2]<0.0){ tmp = tmin[2]; tmin[2] = tmax[2]; tmax[2] = tmp; }

  float mint = MAX3(tmin[0], tmin[1], tmin[2]);
  float maxt = MIN3(tmax[0], tmax[1], tmax[2]);

  // clip p1 and p2 to box
  if (mint>t0) t0 = mint;
  if (maxt<t1) t1 = maxt;

  return;
}

// This code is slow as hell now
bool Isosurface::intersectRay(Ray &r, double t0, double t1) {
  r.t = -1;
  r.prim = NULL;
  r.bc = SbVec3f(1,0,0);

  //if (r.ignorePrim == this) return false;

  const double myEps = 1e-3;//0.001; // tanscale(fov_h/res_h)   // error at 1u
  const double stept = myEps;//1./100;//2048.;//1/256.;//0.01;
  const double maxGrad = 1;
  const double maxGradInv = 1/maxGrad;

  double h0, h1, hm;
  double d0, d1, dm;
  Vec3d p0, p1, pm; // in ONB space

  // intersect ray with slab
  clipToVolume(r, t0, t1);
  
  if (t1<=t0)
    return false;
  
  d0 = t0;
  
  p0 = r.pos+r.dir*d0;
  h0 = height(p0, d0);
  
  // check for already near

//   if (fabs(p[2]-h0)<myEps) {

//     if (DEBUG_STATE)
//       cerr << "hit!\n";
      
//     // hit heightfield
//     r.t = d0;
//     r.prim = this;
//     r.bc = p0;
//     return true;
//   }
  
  bool above = h0>c;

  static bool debuggingEnabled = false;
  if (debuggingEnabled && DEBUG_STATE)
    cerr << "Begin marching\n";

  // iterate
  d1 = d0;
  p1 = p0;
  h1 = h0;
  unsigned int steps = 0;
  while (d1<t1) {
    
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";

    d0 = d1;    // distance
    p0 = p1;    // position (u,v, height)
    h0 = h1;    // heightfield height
    
    d1 += MAX2(d1*myEps, fabs((h1-c)*maxGradInv));           //stept;
    steps++;
    
    if (d1>t1){
      if (debuggingEnabled && DEBUG_STATE)
        cerr << __LINE__ << ": Reached end\n";
      d1 = t1;
    }

    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": d1=" << d1 << "\n";
    
    // get height
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": p1= " << p1 << "\n";
    
    p1 = r.pos+r.dir*d1;
    h1 = height(p1, d1);
    if (debuggingEnabled && DEBUG_STATE)
      cerr << __LINE__ << ": h1=" << h1 << "\n";
    
    // detect crossing
    if (above && h1<=c ||
        !above && h1>=c /*||
        fabs(p1[2]-h1)<myEps*/) {


      if (debuggingEnabled && DEBUG_STATE)
        cerr << "hit... now refine...\n";

      // refine intersection
//      for (int i=0; i<4; i++){
      do {
//        dm = (d0+d1)*0.5;         // bisection method

        double hdiff = h1-h0;
        if (hdiff == 0) break;
        dm = (h1*d0-h0*d1)/hdiff; // false position method
        
        pm = r.pos+r.dir*dm;
        hm = height(pm, dm);

        if (above && hm<c ||
            !above && hm>c /*||
            fabs(pm[2]-hm)<myEps*/) {
          d1=dm;
          p1=pm;
          h1=hm;
        } else {
          d0=dm;
          p0=pm;
          h0=hm;
        }

//      }
//      } while (diff>myEps && h0!=0 && h1!=0);
      } while (fabs(h0-c)>myEps && (d1-d0)>myEps);
        
      // hit heightfield
      r.t = d0; // choose nearer hit, give next ray a better
                // chance of not self-intersecting
      r.prim = this;
      r.bc = p0;
      if (debuggingEnabled && DEBUG_STATE)
        cerr << "returning t=" << d0 << "\n";
      return true;
    }

  }

  if (debuggingEnabled && DEBUG_STATE)
    cerr << __LINE__ << ": no hit\n";
  return false;
}

SbVec3f Isosurface::getNormal(const SbVec3f &bc) const {

  // calculate gradient of scalar field
  const float dp = 0.001;
  const float dpInv = 1/dp;

  double h = height(bc);
  SbVec3f grad(
    (height(Vec3d(bc[0]+dp, bc[1],    bc[2]   )) - h) * dpInv,
    (height(Vec3d(bc[0],    bc[1]+dp, bc[2]   )) - h) * dpInv,
    (height(Vec3d(bc[0],    bc[1],    bc[2]+dp)) - h) * dpInv
    );
  grad.normalize();
  return grad;
}

SbColor Isosurface::getIntDiffColor(const Ray &r) const {
  return SbColor(.5,.5,.5);
//   float v = fmodf(bc[2],10000.)/10000.;
//   SbColor c(v, 0, 0);
//   //cerr << "bc=" << bc << " c=" << c << "\n";
//   return c;
}


// Permutation array as suggested by Steven Worley in
// "A Cellular Texture Basis Function"
class PermArr {

public:
  int arr[256];
  
  PermArr() {
    for (int i=0; i<256; i++)
      arr[i] = i;
    for (int i=0; i<256; i++){
      int j = int(drand48()*256.);
      int t = arr[j];
      arr[j] = arr[i];
      arr[i] = t;
    }
  }

  int val(int i) const {
    return arr[i&255];
  }
  
};

PermArr permArr;

double Isosurface::height(const Vec3d &x_) const {
//  Vec3d x = x_*.1;
  Vec3d x = x_;

  // Helicoid
//   const float a = M_PI/2/10;
//   float b = -x[0]*sin(a*x[1]) + x[2]*cos(a*x[1]);
//   return b;

  // tile floor
  float retDist = 0;
  {
  // 5x5x1 box
  const float gs = .1;       // grid block size
  const float gs2 = gs*2;    // 2x grid block size
  const float gap = .1*gs;   // gap size
  const float ts = gs-gap;   // tile size

  int bi = int( (x[0]-gs)/gs2 );
  int bj = int( (x[2]-gs)/gs2 );
  int bs = bi+bj;

  float p[3];
  p[0] = fabs(x[0]);
  p[1] = fabs(x[1]);
  p[2] = fabs(x[2]);
  
  float f[3];
  f[0] = fmodf(p[0], gs2);
  f[1] = p[1];
  f[2] = fmodf(p[2], gs2);

  if (f[0]>gs)
    f[0] = gs2-f[0];
  
  if (f[2]>gs)
    f[2] = gs2-f[2];

  bs = permArr.val(permArr.val(bi)+bj);
  
  if ( bs%5==0 ) {
    if (f[0]>f[2])
      f[0] = gs-f[0]+gs;
    else
      f[2] = gs-f[2]+gs;
  }
  
  float a = f[0]-ts;
  float b = f[1]-gap;
  float c = f[2]-ts;
  float m = MAX3(a,b,c);
  retDist = 1e5;//m;
  }
/*
  // columns
  {
  const float gs = 2;        // grid block size
  const float gs2 = gs*2;    // 2x grid block size
  const float gap = .9*gs;   // gap size
  const float ts = gs-gap;   // tile size

  int bi = int( (x[0]-gs)/gs2 );
  int bj = int( (x[2]-gs)/gs2 );
  int bs = bi+bj;

  float p[3];
  p[0] = fabs(x[0]);
  p[1] = fabs(x[1]*2-gs);
  p[2] = fabs(x[2]);
  
  float f[3];
  f[0] = fmodf(p[0], gs2);
  f[1] = p[1];
  f[2] = fmodf(p[2], gs2);

  if (f[0]>gs)
    f[0] = gs2-f[0];
  
  if (f[2]>gs)
    f[2] = gs2-f[2];
  
  float a = f[0]-ts;
  float b = f[1]-gs;
  float c = f[2]-ts;
  float m = MAX3(a,b,c);
  
  retDist = MIN2(retDist, m);
  }
  {
  const float gs = 2;         // grid block size
  const float gs2 = gs*2;     // 2x grid block size
  const float gap = .85*gs;   // gap size
  const float ts = gs-gap;    // tile size

  int bi = int( (x[0]-gs)/gs2 );
  int bj = int( (x[2]-gs)/gs2 );
  int bs = bi+bj;

  float p[3];
  p[0] = fabs(x[0]);
  p[1] = fabs(x[1]*6-gs);
  p[2] = fabs(x[2]);
  
  float f[3];
  f[0] = fmodf(p[0], gs2);
  f[1] = p[1];
  f[2] = fmodf(p[2], gs2);

  if (f[0]>gs)
    f[0] = gs2-f[0];
  
  if (f[2]>gs)
    f[2] = gs2-f[2];
  
  float a = f[0]-ts;
  float b = f[1]-gs;
  float c = f[2]-ts;
  float m = MAX3(a,b,c);
  
  retDist = MIN2(retDist, m);
  }
*/
  // ceiling
  {
  const float gs  = 2;        // grid block size
  const float gs2 = gs*2;     // 2x grid block size
  const float gap = .1*gs;    // gap size
  const float ts  = gs-gap;   // tile size

  float p[3];
  p[0] = fabs(x[0]);
  p[1] = x[1];
  p[2] = fabs(x[2]);
  
  float f[3];
  f[0] = fmodf(p[0], gs2);
  f[1] = p[1];
  f[2] = fmodf(p[2], gs2);

  if (f[0]>gs)
    f[0] = gs2-f[0];
  
  if (f[2]>gs)
    f[2] = gs2-f[2];

  float fx = f[0]*f[0] + f[2]*f[2];
//  fx -= gap;
//  fx += gap/2;
  fx *= 10;
  float fy = gs-f[1];
  
  float minv=1e-5;//1/gs;
  if (fx<minv)
    fx = minv;
  if (fy<1e-5)
    fy = 1e-5;
//  fx = MAX2(fx, minv);
//  fy = MAX2(fy, minv);

  float fxd = (fx-1/fy);//-gap*1;
  float fyd = (fy-1/fx);//-gap*1;
  
  fxd/=10;  

  float fd = fxd>0 ?
    MIN2(fxd,fyd) :
    MAX2(fxd,fyd);

  // cut off at y=0 (floor)
//     if (p[1]<0)
//       fd = MAX2(-p[1], fd);
//     else
//       fd = MIN2(-p[1], fd);
  fd = -p[1];

   // thin shell
   float bot = -gap/10;
   if (fd<bot)
     fd = bot-fd;

   // hole
   float tmp = fabs(fx-gs/2);
   fd += 1-tmp;
   
//   if (fd>gap || fd<-gap)
     fd*=.5;

//  return fd;       
  
  retDist = MIN2(retDist, fd);
  }

  return retDist;

//  return x.length()-1; // unit sphere
//  return (x.length()-1) + 0.1*noise.noise(x); // bumpy sphere
//  return x[1] - noise.noise(x); // heightfield

  // spheres at integer corners
//   Vec3d y = x/10;
//   double d = fabs(y[1]);
//   if (d>0.5) return d-0.5;
//   Vec3d corner(floor(y[0]+0.5), floor(y[1]+0.5), floor(y[2]+0.5));
//   return (y-corner).length() - 0.1;

  // blobs on rectilinear grid
//   Vec3d y = x;
//   return noise.noise(y)-0.2;

  // worley noise
  int ii = (int)floor(x[0]);
  int jj = (int)floor(x[1]);
  int kk = (int)floor(x[2]);
  
  SbVec3f xp(x[0]-ii, x[1]-jj, x[2]-kk);
  SbVec3f xp1(1-xp[0], 1-xp[1], 1-xp[2]);
  
  unsigned int seed = 541*ii + 79*jj + 31*kk;
  RNG rng(seed);
  SbVec3f p(rng(), rng(), rng());
  SbVec3f diff = p-xp; 
  float dist2 = diff.dot(diff);
  float dist = sqrt(dist2);
  
  for (int i=-1; i<=1; i++){
    if (i==-1 && dist<xp[i] ||
        i==1 && dist<xp1[i])
      continue;
    for (int j=-1; j<=1; j++){
      if (j==-1 && dist<xp[j] ||
          j==1 && dist<xp1[j])
        continue;
      for (int k=-1; k<=1; k++){
        if (k==-1 && dist<xp[k] ||
            k==1 && dist<xp1[k] ||
            k==kk && i==ii && j==jj)
          continue;

        unsigned int seed = 541*(i+ii) + 79*(j+jj) + 31*(k+kk);
        RNG rng(seed);
        SbVec3f pNew(rng()+i, rng()+j, rng()+k);
        diff = pNew-xp;
        float dist2new = diff.dot(diff);
        if (dist2new<dist2){
          dist2 = dist2new;
          dist = sqrt(dist2);
          p = pNew;
        }
      }
    }
  }

  return dist-0.1;
  
}
double Isosurface::height(const Vec3d &x, float d) const {
  return height(x); // don't use adaptive octaves for now

  // clamp d
  if (d<1) d = 1;

//    static int minOct = 30;
  // from QAEB paper by K. Musgrave, bottom of page 2
//    cerr << __LINE__ << " wtf errorAt1*d=" << errorAt1*d << "\n";
//    int oct = (int)( 0.5 + -3.5-fastlog2(float(errorAt1*d)) );
//     int oct = (int)( 0.5 + -3.5 - log(errorAt1*d)*(1/log(2)) );
//    cerr << __LINE__ << " wtf oct= " << oct << "\n";

//     if (oct<1)
//       oct = 1;
    
//     if (oct>10){
//       cerr << "oct=" << noise.octaves << " d=" << d << " errorAt1=" << errorAt1 <<"\n";
//       exit(1);
//     }

//     if (oct<minOct){
//       minOct = oct;
//       cerr << "minOct = " << minOct << "\n";
//     }

//    noise.octaves = oct;
  return height(x);
}
