#ifndef ISOSURFACE_H
#define ISOSURFACE_H

#include <iostream>
#include <iomanip>
#include "Primitive.h"
#include "Material.h"
#include "Misc.h"
#include "Noise.h"
#include "debug.h"
#include "Main.h"

using namespace std;

class Isosurface : public Primitive {

protected:

  static Material defaultMat;
  float c; // isovalue 
  Material *mat;
  SbVec3f min, max;  // domain of isosurface
  float errorAt1;  // allowable error at d=1 (needed for octaves)
                   // = tan(fov/res)
  mutable Noise noise;
  
public:

  Isosurface(float vertFOV=45, int vertRes=768);
  
  virtual PrimType getType() const { return PRIMTYPE_ISOSURFACE; }
  virtual string getTypeName() const { return "Isosurface"; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);

  // be sure to initialize t0, t1 to some values first!
  void clipToVolume(const Ray &r, double &t0, double &t1);

  SbVec3f getMin() const {
    return min;
  }
  SbVec3f getMax() const {
    return max;
  }

  Material* getMat() { return mat; }
  const Material* getMat() const { return mat; }
  SbVec3f getNormal(const SbVec3f &bc)  const;

  static Material* getDefaultMat() { return &defaultMat; }

  // testing purposes
  SbColor getIntDiffColor(const Ray &r) const;

  double height(const Vec3d &x) const;
  double height(const Vec3d &x, float d) const;
};

#endif
