#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include "Julia.h"
#include "RNG.h"

Material Julia::defaultMat;

double Julia::dist(const Quat &z_) const {
  Quat z = z_, zp = Quat(1,0,0,0), zp_next, z_next;
  double zn2_next=0, zn2=0;

  const int n = 30;
  for (int i=0; i<n; i++){

    // TODO:
    // adjust epsilon depending on dist from eye

    // compute z(n+1), zp(n+1)
    zp_next = 2*(z*zp);
    z_next = z*z + fc;
    
    // dz = z.len()/(2*zp.len())*log(z.len());
    // but this is computed below

    // detect impending overflow
    zn2_next = z_next.len2();
    if (zn2>200000.*200000.) break;
    
    zp = zp_next;
    z = z_next;
    zn2 = zn2_next;
  }

  // subst x=sqrt(x^2) into 0.5*|z|/|z'|*log|z|
  return sqrt(zn2/zp.len2())*0.25*log(zn2); 
}


void Julia::initBasis(int seed){

  RNG rng(seed);
  
  // pick random ortho normal basis
  // using gram schmidt for orthonormalization
  aB = Quat(rng(), rng(), rng(), rng()).norm();
  do {
    bB = Quat(rng(), rng(), rng(), rng()).norm();
  } while (bB.dot(aB)<0.1);
  bB = (bB-bB.dot(aB)*aB).norm();
  do {
    cB = Quat(rng(), rng(), rng(), rng());
  } while (cB.dot(aB)<0.1 || cB.dot(bB)<0.1);
  cB = (cB-cB.dot(aB)*aB).norm();
  cB = (cB-cB.dot(bB)*bB).norm();
}

// intersect julia set
bool Julia::intersectRay(Ray &r, double t0, double t1) {
  
  // test bounding sphere
  double t, near = 0, far = 0;

  r.t = -1;
  r.prim = NULL;
  //r.bc = SbVec3f(1,0,0);
  
  if (juliaSphere.intersect(r, -juliaSphere.rad, t, far)) {

    Quat curp;
    double curd = 0;
    const double lim = 0.5;

    // Transform ray to quaternion space
    Quat ro = worldVecToQuat(r.pos);
    Quat rd = worldDirToQuat(r.dir);

    if (far>t1)
      far = t1;
    
    t *= invjScale;
    t0 *= invjScale;
    far *= invjScale;

    if (r.ignorePrim == this && t0<twojeps)
      t0 = twojeps;
    else if (t0<jeps)
      t0 = jeps;
    
    if (t<t0)
      t = t0;
    
    // step (t+=d) until d<eps
    do {
      
      t += curd;
      
      curp = ro + rd*t;
      curd = dist(curp);
      if (curd>lim) curd = lim;
      
    } while (curd>jeps && t<far);
    
    if (curd<=jeps){
      t = t-(jeps-curd); // back t up to isosurface
      r.t = t*jScale;
      r.prim = this;
      r.bc = r.pos + r.t*r.dir;
      return true;
    }
  }
  
  // no hit
  return false;
}
