#include <stdlib.h>
#include <alloca.h>
#include "Pane.h"
#include "debug.h"
#include "DataSet.h"

using namespace std;


//'''''''''''WW'''''''WW''WW''''WW''''''WW''''WW'''WWW''''''''''''''WW'''''''
//                    WW  WW            WW    WW  WW  W             WW       
//WWW   WW   WW WWW  WWWW WW    WW  WWW WWW  WWWW WW     WW   WWW  WWWW W WW 
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW    WW W  WW W  WW  WW   
//WW W WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW  W WW W  WW W  WW  WW   
//WWW   WW   WW WW W  WW  WWWWW WW  WWW WW W  WW   WWW   WW   WW W  WW  WW   
//WW                                 WW                                      
//WW'''''''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''
//
// Calculates point lights contribution (does not require surface)
//
SbColor Pane::pointLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng){
  
  SbColor plcolor = SbColor(0.0, 0.0, 0.0);
  float value;
  SbVec3f l;
  vector<SoPointLight *>::iterator pI = litPnts.begin();
  if (gv->headLightPos == HL_OFF)
    pI++;

  for (; pI != litPnts.end(); pI++){

    // compute light direction, l
    l = (*pI)->location.getValue() - r.x;
    float dist = l.normalize();

    // dont use lights below horizon
    if (ng.dot(l) <= 0.0)
      continue;

    // don't illuminate areas that think they are facing away
    float dot1 = n.dot(l);
    if (dot1<=0.0)
      continue;
    
    // trace shadow ray
    //Ray sr((*pI)->location.getValue(), -l);
    // Ray sr(r.x+0.1*l, l, r.ctx); // shadow offset improves things
    Ray sr(r.x, l, r.ctx, r.prim);
    //sr.show2(dist);

    // does light ray reach triangle?
    const float t_eps = 0.002;
    if (scenePrim->intersectRaySimple(sr, t_eps) && sr.t<dist)
      continue;

    sr.t = dist;

    // from jensen p.36 eq. 3.2 (frd = diffuse brdf, f_r,d)
    // L = frd*(w.n)/dist^2*Phi_l/(4pi)
    value = (*pI)->intensity.getValue() /
      ( gv->meters2(sr.t*sr.t) * (4.0*M_PI) );
    // (l.n)/dist^2 * 1/(4pi)

    // compute color
    // from jensen p.36 eq. 3.2

    Material *mtl = r.prim->getMat(r);
    BRDF *brdf = mtl->getBRDF();
    SbVec3f brdfCosT;
    float pdfBrdf;
    brdf->GetBRDF(r.dir, l, n, ng, r, brdfCosT, pdfBrdf);

#ifdef SCHLICK_CYLINDER_TEST    
    value = 1;
//     float c = n.dot(l);
//     brdfCosT = SbVec3f(c,c,c);
//    brdfCosT *= M_PI;
#endif
    
    SbColor lcolor = (*pI)->color.getValue(); // in W
    SbVec3f lx = (*pI)->location.getValue();
    gv->pane->applyAttenuation(r, r.x, lx, dist, lcolor);

    plcolor += vecScale(brdfCosT*value, lcolor);
    
    // frd * [ (l.n)/dist^2 * 1/(4pi) ] * Phi_l        note: frd = p/pi
  }
  
  return plcolor;
}




//'''''''''''WWW''WWWW''''''''WW'''''''WW''WW''''WW''''''WW''''WW'''WWW'''''''
//            WW  WW  W                WW  WW            WW    WW  WW  W      
//WW W  WW    WW  WW  W  WW   WW WWW  WWWW WW    WW  WWW WWW  WWWW WW     WW  
//WW W WW W   WW  WWWW  WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW    WW W 
//WW W WW W   WW  WW    WW W  WW WW W  WW  WW    WW W WW WW W  WW  WW  W WW W 
// WW   WW   WWWW WW     WW   WW WW W  WW  WWWWW WW  WWW WW W  WW   WWW   WW  
//                                                    WW                      
//''''''''''''''''''''''''''''''''''''''''''''''''''WWW'''''''''''''''''''''''
// Experimental volume scattering code
//
SbColor Pane::volPointLightContrib(const Ray &r){

  // find contribution of point lights
  SbColor plcolor = SbColor(0.0, 0.0, 0.0);
  float value;
  SbVec3f l;
  vector<SoPointLight *>::iterator pI = litPnts.begin();
  if (gv->headLightPos == HL_OFF)
    pI++;

  for (; pI != litPnts.end(); pI++){
      
    // compute light direction, l
    l = (*pI)->location.getValue() - r.x;
    float dist = l.normalize();
      
    // trace shadow ray
    Ray sr(r.x, l, r.ctx, r.prim);
      
    // does light ray reach triangle?
    if (scenePrim->intersectRay(sr) && sr.t<dist)
      continue;
    sr.t = dist;
      
    // from jensen p.36 eq. 3.2 (frd = diffuse brdf, f_r,d)
    // L = frd*(w.n)/dist^2*Phi_l/(4pi)
    value = 1.0 / gv->meters2(sr.t*sr.t) /
      (4.0*M_PI) * (*pI)->intensity.getValue();
    // (l.n)/dist^2 * 1/(4pi)

    // no absorbtion
    // value *= exp(-sigma_t*gv->meters(sr.t));

    // compute color
    // from jensen p.36 eq. 3.2
    SbColor lcolor = (*pI)->color.getValue(); // in W
    SbVec3f lx = (*pI)->location.getValue();
    gv->pane->applyAttenuation(r, r.x, lx, dist, lcolor);
    
    plcolor += lcolor*value;
    // frd * [ (l.n)/dist^2 * 1/(4pi) ] * Phi_l        note: frd = p/pi
  }

  return plcolor;
}


//'''''''''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''''WW''WWW'
//                 W          W    W  W  W          W        W   W  W W  W
// WW  WW  WW   WW W    W  WW WW  WWW W     W  WW  WWW  WW W WW    W  W  W
//W W W   Wmmm W W W    W W W W W  W  W  W W W W W  W  W   W W W  W   W  W
// WW W    WW   WW WWWW W  WW W W  W   WW   W  W W  W  W   W WW  WWWW WWW 
//                          W                                             
//                        WW                                              
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution in 2 Dimensions
//
SbColor Pane::areaLightContrib2D(Ray &r, SbVec3f n, SbVec3f ng){

  float value;
  double cA, cB, cC;
  SbVec3f l;
  SbVec3f lx;
    
  SbColor alcolor = SbColor(0.0, 0.0, 0.0);

  // fix inward pointing normals
  if (n.dot(ng)<0.0)
    n = -n;

  // flatten normals
  n[1] = 0.0;
  n.normalize();
  ng[1] = 0.0;
  ng.normalize();

  // do nothing if no light samples
  if (gv->lightSamples<=0)
    return(alcolor); // black
  
  // loop over area lights
  for (int ai=0; ai< litTris.size(); ai++){

    // intersect light with plane
    EmissiveTriangle *tri = &litTris[ai];    
    if (tri == r.ignorePrim) continue;
    SbVec3f c[2];
    int crossing=0;

    // loop over vertices
    for (int i=0; i<3; i++){
      int j = (i+1)%3;
      if (crossing>2){
	cerr << "crossing error!\n";
	continue;
      }

      if ( (tri->vert(i)[1]-r.x[1])*(tri->vert(j)[1]-r.x[1]) < 0 ){
	float ratio = (r.x[1] - tri->vert(i)[1]) /
	  (tri->vert(j)[1] - tri->vert(i)[1]);
	c[crossing][0] = ratio * (tri->vert(j)[0] - tri->vert(i)[0]) +
	                 tri->vert(i)[0];
	c[crossing][1] = r.x[1];
	c[crossing][2] = ratio * (tri->vert(j)[2] - tri->vert(i)[2]) +
	                 tri->vert(i)[2];
	crossing++;
      }
    }

    // debugging
    // if (crossing){
    //   showPoint(c[0], 3.0);
    //   showPoint(c[1], 3.0);
    //   showPoint(tri->vert[0], 1.0);
    //   showPoint(tri->vert[1], 1.0);
    //   showPoint(tri->vert[2], 1.0);
    // }

    // Loop over samples
    SbColor colora = SbColor(0.0, 0.0, 0.0);
    for (int lsamp=0; lsamp<gv->lightSamples; lsamp++){

      if (gv->areaLights){
	// choose random position inside light
	cA = r.ctx->mydrand48();
      }
      else
	cA = 0.5;

      // compute sample location
      lx = cA*c[0] + (1.0-cA)*c[1];

      // compute light direction, l
      l = lx - r.x;
      l[1] = 0.0;
      float lDist = l.normalize();

      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0)
        continue;
	
      // trace from point to light point
      Ray lr = Ray(r.x, l, r.ctx, r.prim);
      if (scenePrim->intersectRay(lr, 0.002, lDist) && lr.prim != tri)
	continue; // does not hit light

      lr.t = lDist;

      // don't illuminate area that think they are facing away
      double dot1 = n.dot(l);
      if (dot1<0.0)
	continue;

      // calculate light normal scalar
      SbVec3f bc;
      tri->calcBaryCoords(lx, bc);
      SbVec3f np = tri->getIntNormal(bc);
      np[1] = 0.0;
      np.normalize();
      double dot2 = -np.dot(l);
      if (dot2<0.0)
	continue;

      // compute value (fraction of light reaching surface)
      // from shirley p.120 eq. 15.5
      // Le = Phi-l/area/pi
      // L(x) = (emissive Le+) p/pi*Le(x')*cos(theta1)*area*cos(theta2)/dist^2
      value = dot1 * dot2 * gv->meters( (c[1]-c[0]).length() ) /
	gv->meters(lr.t);
      // (cos(theta)*area*cos(theta')/dist^2

      // from shirley p.124 middle of page
      SbColor Le = tri->getEmission() / // in W
	gv->meters2( tri->area ) / 2.0;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light
		   
      // compute color
      // shirley p.120 eq. 15.5
      colora += SbColor(r.diffCol[R]/2.0 * Le[R] * value,
			r.diffCol[G]/2.0 * Le[G] * value,
			r.diffCol[B]/2.0 * Le[B] * value);
    } // end for each sample

    alcolor += colora/gv->lightSamples;
      
  } // end for each light

  return(alcolor);
}


//'''''''''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''
//                 W          W    W  W  W          W        W  
// WW  WW  WW   WW W    W  WW WW  WWW W     W  WW  WWW  WW W WW 
//W W W   Wmmm W W W    W W W W W  W  W  W W W W W  W  W   W W W
// WW W    WW   WW WWWW W  WW W W  W   WW   W  W W  W  W   W WW 
//                          W                                   
//                        WW                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution
// Samples every light. Less efficient than
// than the importance sampling method (below).
//
SbColor Pane::areaLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng){

  float value;
  double cA, cB, cC;
  SbVec3f l;
  SbVec3f lx;
    
  SbColor alcolor = SbColor(0.0, 0.0, 0.0);

  // no longer necessary - caller is responsible
  // for passing upward facing normal
  //
  // fix inward pointing normals
  // if (n.dot(ng)<0.0)
  //   n = -n;

  // do nothing if no light samples
  if (gv->lightSamples<=0)
    return(alcolor); // black

  // loop over area lights
  // TODO: Thank Chris Weigle for the <= bug fix below!
  for (int ai=0; ai<litTris.size(); ai++){

    EmissiveTriangle *tri = &litTris[ai];
    
    SbColor colora = SbColor(0.0, 0.0, 0.0);
    for (int lsamp=0; lsamp<gv->lightSamples; lsamp++){

      if (gv->areaLights){
	// choose random position inside light
	cB = 1.0-sqrt(r.ctx->mydrand48());
	cC = (1.0-cB)*r.ctx->mydrand48();
	cA = 1.0-cB-cC;
      }
      else
	cB = cC = cA = 1.0/3.0;
	
      lx = cA*tri->vert(0) + cB*tri->vert(1) + cC*tri->vert(2);

      // compute light direction, l
      l = lx - r.x;
      
      // 2d adjustment
      if (gv->useDimensions<3)
	l[1] = 0.0;
      
      float lrt = l.normalize();

      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0)
        continue;
	
      // trace from point to light point
      Ray lr = Ray(r.x, l, r.ctx, r.prim);
      if ( scenePrim->intersectRay(lr, 0.002, lrt) && lr.prim != tri)
        continue; // does not hit light

      lr.t = lrt;

      // don't illuminate area that think they are facing away
      double dot1 = n.dot(l);
      if (dot1<0.0)
	continue;

      // calculate light normal scalar
      SbVec3f bc(cA, cB, cC);
      SbVec3f np = tri->getIntNormal(bc);
      double dot2 = -np.dot(l);
      if (dot2<0.0)
	continue;

      // compute value (fraction of light reaching surface)
      // from shirley p.120 eq. 15.5
      // Le = Phi-l/area/pi
      // L(x) = (emissive Le+) p/pi*Le(x')*cos(theta1)*area*cos(theta2)/dist^2
      value = dot1 * dot2 * gv->meters2( tri->area ) /
	gv->meters2(lr.t*lr.t);
      // (cos(theta)*area*cos(theta')/dist^2

      // from shirley p.124 middle of page
      SbColor Le = tri->getEmission() / // in W
	gv->meters2( tri->area ) * M_1_PI;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light
		   
      // compute color
      // shirley p.120 eq. 15.5
      colora += SbColor(r.diffCol[R]*M_1_PI * Le[R] * value,
			r.diffCol[G]*M_1_PI * Le[G] * value,
			r.diffCol[B]*M_1_PI * Le[B] * value);
    } // end for each sample

    alcolor += colora/gv->lightSamples;
      
  } // end for each light

  return(alcolor);
}


//'''''''''''''''''WWW'''''''''''W''''m'''''W''''''''WW''''''''''''''''''m'W''
//                  W            W          W    W  W  W          W        W  
// WW  WW  WW   WW  W  WW W  WW  W    W  WW WW  WWW W     W  WW  WWW  WW W WW 
//W W W   Wmmm W W  W  W W W W W W    W W W W W  W  W  W W W W W  W  W   W W W
// WW W    WW   WW WWW W W W WW  WWWW W  WW W W  W   WW   W  W W  W  W   W WW 
//                           W            W                                   
//                           W          WW                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Calculates area lights contribution
// Uses multiple luminaire importance sampling
// Discussed in Peter Shirleys Realistic Ray Tracing 1st ed. p. 124
// Basically, we simulate a discrete random variable
// to sample a probability mass function, where the masses
// are proportional to the importance (proximity and direction)
// of the lights.
//
SbColor Pane::areaImpLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, int simple){

  if (gv->useDimensions==2)
    return(areaLightContrib2D(r, n, ng));
  
  SbColor color = SbColor(0.0, 0.0, 0.0);
  SbVec3f l, xP, xPP, nP;
  EmissiveTriangle *lTri;

  // the following is unnecessary because the caller has already
  // given us a nice upward facing normal (it had better!)
  // fix inward pointing normals
  // if (n.dot(ng)<0.0)
  //   n = -n;

  int numLights = litTris.size();
  if (numLights == 0 || gv->lightSamples <= 0)
    return color; // black

  double cA, cB, cC;
  double e_i, dot1, dot2, dot3;
  double lP, lP2, lPP;

  // initialize precomputation tables
  typedef struct cacheEntry {
    float sarea, alpha, cos_alpha, sin_alpha;
    float product; // sin-alpha*cos_c
    SbVec3f CgA;
  } cacheEntry_t;

  enum sampType_t { APPROX_SAMPLING, ACCURATE_SAMPLING };
  float *weights = (float *)alloca(sizeof(float)*numLights);
  sampType_t *sampleMethod = (sampType_t *)alloca(sizeof(sampType_t)*numLights);   // 0=approx, 1=acc
  cacheEntry_t *cv = (cacheEntry_t *)alloca(sizeof(cacheEntry_t)*numLights);
  unsigned int *cvComputed = (unsigned int *)alloca(sizeof(unsigned int)*numLights);

  /////////////////////////
  // compute weights[]
  /////////////////////////
  double eSum = 0;
  int lNum = 0;
  int nonZeroWeights = 0;
  for (int ai=0; ai < litTris.size(); ai++,lNum++){
    
    // L*cosO*cosO'*A/dist^2
    
    cvComputed[lNum] = 0;

    // compute Lavg
    lTri = &litTris[ai];
    e_i = lTri->getEmisAvg()/lTri->area;  // ~ L(i)

    // Figure purposes - only allow certain lights
    if (gv->figureMode && gv->useGlassModel==0 &&
	ABSF(e_i-0.15)>0.01)
      continue;
    
    // find normal
    nP = lTri->getNormalPre();

    //                      /\
    //                     /  \
    //                    /    \
    // 	     nP<------xP \
    //                  '---/----'
    // 		    /
    //           nPP      /
    //            ^\  p3 /l
    //              \ /\/        
    //               \ /\
    //              / xPP\
    //           p2'-/----p1
    //              x

    // find center
    const SbVec3f vert[3] = { lTri->vert(X), lTri->vert(Y), lTri->vert(Z) };
    
    xP = (vert[X]+vert[Y]+vert[Z]) * (1./3.);

    // make rough estimate
    l = xP-r.x;
    lP = l.normalize();

    // calculate dot products and cos's
    float dp[3], dp1[3];
    int dpcnt = 0;
    dot1 = 0.0;
    for (int i=0; i<3; i++){
      xP = vert[i]-r.x;
      dp[i] = ng.dot(xP);         //  dp = ||1|| ||d|| cos(theta)
      dp1[i] = dp[i]/xP.length(); // dp1 = cos(theta)

      // calculate average angle between normal and light
      if (dp1[i]>0.0){
        dot1 += dp1[i];
        dpcnt++;
      }
    }

    // check luminaire is above horizon
    //if (dot1 <= 0.0){
    if (dpcnt == 0){
      weights[lNum] = 0.0;
      continue;
    }

    dot1 /= dpcnt;  // can be delayed, for speed
    
    // calculate angle of light normal and light vector
    dot2 = -nP.dot(l);
    
    // check to see if luminaire pointing away
    if (dot2 <= 0.0){
      weights[lNum] = 0.0;
      continue;
    }

    // calculate projected solid angle (max pi, area of unit circle)
    float w1 = lTri->area / (lP*lP); // max projected area
    float w2 = dot1 * dot2;          // approximate projected area scalar components

    // if relatively far away (max solid angle is small)
    // or miniscule solid angle -> approximate sampling
    // if partially below horizon and not very small
    // otherwise -> accurate

    // weight
    //
    //
    // sampling
    // approximate if w_ < 0.001 || w1 < pi/20
    // acccurate if w_ > 0.001

    // if maximum solid angle is very small,
    // or if at extreme angle, use approximate weighting & sampling 
    if (w1 < 0.001 ||
        w2 < 0.001){
      
      // use approximate sampling
      sampleMethod[lNum] = APPROX_SAMPLING;

      e_i *= w1 * w2;

      //return(SbColor(1,0,0));
    }
    else {    
      //return(SbColor(0,1,0));

      // use accurate sampling
      sampleMethod[lNum] = ACCURATE_SAMPLING;

      // temporary bug fix
      //weights[lNum] = 1.0;
      //eSum += 1.0;
      //continue;

      // compute projected solid area using Stoke's thm
      // from Improving Radiosity Solutions
      // through the Use of Analytically Determined
      // Form Factors by Baum, Rushmeier, and Winget
      // (Eq. 9 on pg. 6 (or "330"))

      SbVec3f pts[4];
      int ptscnt = 0;

      // clip triangle against horizon
      static const int nextIdx[] = { 1, 2, 0 };  // j = (i+1)%3
      for (int i=0; i<3; i++){
        int j = nextIdx[i];
        if (dp[i]>=0.0)      // check vertex (don't clip pts on horizon)
          pts[ptscnt++] = vert[i];
        if (dp[i]*dp[j]<0.0) // clip crossing edge against horizon
          pts[ptscnt++] = (vert[i]*(-dp[j]) + vert[j]*dp[i])/(dp[i]-dp[j]);
      }

      // guaranteed to be 3-4 points
      // integrate over edges
      float projarea = 0.0;
      for (int i=0; i<ptscnt; i++){
        int j = (i+1)%ptscnt;
        SbVec3f r1 = pts[i]-r.x;
        SbVec3f r2 = pts[j]-r.x;
        SbVec3f tau = r1.cross(r2);
        r1.normalize();
        r2.normalize();
        float dotp = clamp(r1.dot(r2), -1.f, 1.f);
	
        float gamma = acos(dotp);
        if (!finite(gamma)){
          cerr << "gamma not finite" << endl;
          exit(1);
        }
      
        tau.normalize();
        tau *= gamma;
        projarea -= ng.dot(tau);
	
        if (!finite(projarea)){
          cerr << "problem with projarea\n";
          exit(1);
        }
	
      }

      //projarea /= 2.0*M_PI;  // This would give integral of 1...
      //projarea *= M_PI;      // ...but we want pi
      projarea *= 0.5;         // so I think 1/2 is the normalization we want for the integral
      
      e_i *= projarea;

      // debugging
      //if (gv->tempInt == 39)
      //  e_i = 1;

    }

    weights[lNum] = e_i;
    
    if (e_i){
      nonZeroWeights++;
      eSum += e_i;
    }
  }

  if (eSum == 0.0)
    return color; // black
  
  // add small amount to weights to prevent really small weights
  // (which otherwise would result in high variance in light samples)
  double dw = eSum/(nonZeroWeights*nonZeroWeights);
  eSum += dw*nonZeroWeights;
  
  // debugging
  //if (gv->tempInt == 40)
  //  return(SbColor(eSum, eSum, eSum));
  
  double eSumInv = 1/eSum;
  for (lNum=0; lNum<numLights; lNum++){
    if (weights[lNum]){
      weights[lNum] += dw;
      weights[lNum] *= eSumInv;
    }
//     if (DEBUG_STATE)
//       cerr << "weights [" << lNum << "] = " << weights[lNum] << "\n";
  }

  // debugging
  //if (gv->tempInt == 41)
  //  return(SbColor(weights[0], weights[1], 0.0));
  
  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  // stratified sampling
  int sqrt_samples = (int)sqrt((float)gv->lightSamples);
  
  // use one sample in final gather
  if (simple)
    sqrt_samples = 1;

  double invSqrtSamples = 1./sqrt_samples;

  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) * invSqrtSamples;
      double eps2 = (sampy + r.ctx->mydrand48()) * invSqrtSamples;
      //double eps1 = drand48(), eps2 = drand48();
      double eps2orig = eps2;

      Material *mtl = r.prim->getMat(r);
      BRDF *brdf = mtl->getBRDF();
      SbVec3f brdfCosT;
      float newPdf, pdfLight;
      SbVec3f p1, p2, p3;
      double value=1;

      // i add a small amount to make it not intersect overlapping surfaces
      const float t_eps = 0.002;
      Ray lr = Ray(r.x, l, r.ctx, r.prim);


      
      // Multiple Importance Sampling (MIS)
      float lightWeight = mtl->getGlossiness() ? 0.5 : 1;
      float brdfWeight = 1 - lightWeight;
      if (1<=lightWeight || r.ctx->mydrand48()<lightWeight){
      
      // pick light
      // could be faster with bisectional search instead of linear
      eSum = 0.0;
      double eSumNext = 0.0;
      int lastGood = 0;
      for (lNum=0; lNum<numLights-1; lNum++){
	if (weights[lNum]<=0) continue;
	lastGood = lNum;
	eSumNext += weights[lNum];
	if (eps1<=eSumNext) break;
	eSum = eSumNext;
      }
      if (weights[lNum]<=0.0) lNum = lastGood;

      //if (DEBUG_STATE)
      //cerr << "chose light " << lNum << endl;
      eps1 = (eps1 - eSum) / weights[lNum];

      lTri = &litTris[lNum];

      // Generate sample on solid angle of triangle
      // from "Stratified Sampling of Spherical Triangles"
      // by James Arvo
      // Code mostly glibbed from Java version

      // this would be better if it picked
      // a point on the CLIPPED luminaire
      // oh well, future work
           
      // 1. find p1, p2, p3, points on unit sphere surrounding x
      p1 = lTri->vert(0)-r.x;
      p1.normalize();
      p2 = lTri->vert(1)-r.x;
      p2.normalize();
      p3 = lTri->vert(2)-r.x;
      p3.normalize();

      if (sampleMethod[lNum] == ACCURATE_SAMPLING){

	//return(SbColor(0,1,0));

	// Only compute certain values once and store in
	// a cache. Then use cache values in future computations.
	if (cvComputed[lNum] == 0){
	
	  cvComputed[lNum] = 1;

	  // Compute dihedral angles
	  SbVec3f u12, u13, u23;
	  u12 = p1.cross(p2); u12.normalize();
	  u13 = p1.cross(p3); u13.normalize();
	  u23 = p2.cross(p3); u23.normalize();

	  cv[lNum].cos_alpha =   u13.dot(u12);  // p3xp1 . p2xp1
	  float cos_beta  =    -(u12.dot(u23)); // p1xp2 . p3xp2
	  float cos_gamma =      u13.dot(u23);  // p1xp3 . p2xp3

	  // if far enough away, these points could be
	  // close together
	  if (cv[lNum].cos_alpha > 1.0)
	    cv[lNum].cos_alpha = 1.0;
	  if (cos_beta > 1.0)
	    cos_beta = 1.0;
	  if (cos_gamma > 1.0)
	    cos_gamma = 1.0;
	    

	  cv[lNum].alpha = acos(cv[lNum].cos_alpha);
	  float beta  = acos(cos_beta);
	  float gamma = acos(cos_gamma);

	  cv[lNum].sarea = cv[lNum].alpha+beta+gamma - M_PI;

	  // debugging
	  if (!finite(cv[lNum].sarea))
	    cerr << "problem2\n";
	
	  // compute other cache values
	  cv[lNum].sin_alpha = sin(cv[lNum].alpha);
	  float cos_c = p1.dot(p2);
	  cv[lNum].product = cv[lNum].sin_alpha*cos_c;
	  cv[lNum].CgA = p3-p1*(p3.dot(p1)/(p1.dot(p1)));
	  cv[lNum].CgA.normalize();
	}

	// temporary bug fix 
	if (cv[lNum].sarea <= 1.0e-6)
	  continue;

	// find sample 
      
	float phi = eps1*cv[lNum].sarea - cv[lNum].alpha;
	float s = sin(phi);
	float t = cos(phi);
	float u = t-cv[lNum].cos_alpha;

	float v = s+cv[lNum].product;
	float q =
	  (cv[lNum].cos_alpha * (v*t-u*s) - v) /
	  (cv[lNum].sin_alpha * (u*t+v*s)    );
	double temp = MAX2(1.0-q*q, 0.0);

	SbVec3f T = q*p1 + sqrt(temp)*cv[lNum].CgA;
	T.normalize();
        float TdotP2 = T.dot(p2);
        if (TdotP2>1.f) TdotP2 = 1.f;
	float z = 1.f-eps2 * (1.f-TdotP2);
	T -= p2 * (TdotP2/p2.dot(p2));
	T.normalize();
	l = z*p2 + sqrt(1.0-z*z)*T;
      }
      else {

	//return(SbColor(1,0,0));

	// choose sample point using flat triangle

	// 1. find p1, p2, p3, points on unit sphere surrounding x
	p1 += r.x;
	p2 += r.x;
	p3 += r.x;

	cB = 1.0-sqrt(1.0-eps2);
	cC = (1.0-cB)*eps1;
	cA = 1.0-cB-cC;

	xPP = cA*p1 + cB*p2 + cC*p3;
    
	// 4. Find light direction, l
	l = xPP - r.x;
      }

      // 2d adjustment
      if (gv->useDimensions<3)
        l[1] = 0.0;
        
      lPP = l.normalize();
    
      // dont illuminate lights below horizon
      if (ng.dot(l) <= 0.0){
	continue;
      }

      // don't illuminate area that think they are facing away
      dot1 = n.dot(l);
      if (dot1<=0.0)
	continue;      

      // trace from point to light point
      
      // get distance to triangle
      lr.dir = l;
//       if (!lTri->intersectRaySimple(lr, t_eps, 1e8))   // i think this is redundant
// 	continue; // occluded
      
      // use light sampling
      if (!scenePrim->intersectRaySimple(lr, t_eps, 1e8) || lr.prim != lTri)
        continue; // occluded

      // NOTE: we must intersect the light to get lr.t,
      // (for APPROX_SAMPLING and ray marching)
            
      if (sampleMethod[lNum] == ACCURATE_SAMPLING){
	pdfLight = 1/cv[lNum].sarea;
      }
      else {

	// calculate edges and normal
	SbVec3f edge1 = p2-p1;
	SbVec3f edge2 = p3-p1;
	SbVec3f nPP = edge1.cross(edge2);

	double area = 0.5*nPP.normalize();

	nP = lTri->getNormalPre();
	dot2 = -nP.dot(l);

        // light may point orthogonal to point
        if (dot2 <= 0)
          continue;
	
	double dot3 = -nPP.dot(l);
	if (dot3<=0.0)
	  continue;
    
	// 6. Find p(x) =                  ||x-x"||^2 * cosO'
	//                 -----------------------------------
	//                  area(p1p2p3) * ||x-x'||^2 * cosO"

	lP = lr.t;
	double prob = (gv->meters2(lPP*lPP) * dot2) /
	  (gv->meters2(area) * gv->meters2(lP*lP) * dot3);

        pdfLight = (gv->meters2(lP*lP) * prob)/dot2;
      }

      //value *= M_1_PI * dot1;      // diffuse brdf * cos(theta)
      //value *= brdfCosT;

        float pdfBrdf;
        brdf->GetBRDF(r.dir, l, n, ng, r, brdfCosT, pdfBrdf);
        newPdf = lightWeight*pdfLight + brdfWeight*pdfBrdf;
        value /= weights[lNum];

      } else {

        // use brdf sampling
        double brdfVal, pdfBrdf, brdfOverPdfVal, brdfOverPdfCosTval;
        SbVec3f brdfColor;
        double eps3 = r.ctx->mydrand48();
        brdf->GetSampleDir(eps1, eps2orig, eps3, r.dir, n, ng, r, l,
                           brdfVal, pdfBrdf, brdfOverPdfVal, brdfOverPdfCosTval,
                           brdfColor);

        if (pdfBrdf<=0)
          continue;

        dot1 = n.dot(l);
        if (dot1<=0.0)
          continue;      
        
        lr.dir = l;
        if (!scenePrim->intersectRaySimple(lr, t_eps, 1e8) ||
            !lr.prim->isOfType(Primitive::PRIMTYPE_EMISTRIANGLE))
          continue; // occluded

        lTri = (EmissiveTriangle*)lr.prim;
	nP = lTri->getNormalPre();
	dot2 = -nP.dot(l);
        if (dot2<=0) continue;


        // damn... have to compute pdf of my busted sampling
        lNum = lTri - &litTris[0];




      p1 = lTri->vert(0)-r.x;
      p1.normalize();
      p2 = lTri->vert(1)-r.x;
      p2.normalize();
      p3 = lTri->vert(2)-r.x;
      p3.normalize();

      if (sampleMethod[lNum] == ACCURATE_SAMPLING){

	//return(SbColor(0,1,0));

	// Only compute certain values once and store in
	// a cache. Then use cache values in future computations.
	if (cvComputed[lNum] == 0){
	
	  cvComputed[lNum] = 1;

	  // Compute dihedral angles
	  SbVec3f u12, u13, u23;
	  u12 = p1.cross(p2); u12.normalize();
	  u13 = p1.cross(p3); u13.normalize();
	  u23 = p2.cross(p3); u23.normalize();

	  cv[lNum].cos_alpha =   u13.dot(u12);  // p3xp1 . p2xp1
	  float cos_beta  =    -(u12.dot(u23)); // p1xp2 . p3xp2
	  float cos_gamma =      u13.dot(u23);  // p1xp3 . p2xp3

	  // if far enough away, these points could be
	  // close together
	  if (cv[lNum].cos_alpha > 1.0)
	    cv[lNum].cos_alpha = 1.0;
	  if (cos_beta > 1.0)
	    cos_beta = 1.0;
	  if (cos_gamma > 1.0)
	    cos_gamma = 1.0;
	    

	  cv[lNum].alpha = acos(cv[lNum].cos_alpha);
	  float beta  = acos(cos_beta);
	  float gamma = acos(cos_gamma);

	  cv[lNum].sarea = cv[lNum].alpha+beta+gamma - M_PI;

// 	  // debugging
// 	  if (!finite(cv[lNum].sarea))
// 	    cerr << "problem2b\n";
	
// 	  // compute other cache values
// 	  cv[lNum].sin_alpha = sin(cv[lNum].alpha);
// 	  float cos_c = p1.dot(p2);
// 	  cv[lNum].product = cv[lNum].sin_alpha*cos_c;
// 	  cv[lNum].CgA = p3-p1*(p3.dot(p1)/(p1.dot(p1)));
// 	  cv[lNum].CgA.normalize();
	}

	// temporary bug fix 
	if (cv[lNum].sarea <= 1.0e-6)
	  continue;

	// find sample 
      
// 	float phi = eps1*cv[lNum].sarea - cv[lNum].alpha;
// 	float s = sin(phi);
// 	float t = cos(phi);
// 	float u = t-cv[lNum].cos_alpha;

// 	float v = s+cv[lNum].product;
// 	float q =
// 	  (cv[lNum].cos_alpha * (v*t-u*s) - v) /
// 	  (cv[lNum].sin_alpha * (u*t+v*s)    );
// 	double temp = MAX2(1.0-q*q, 0.0);

// 	SbVec3f T = q*p1 + sqrt(temp)*cv[lNum].CgA;
// 	T.normalize();
// 	float z = 1.0-eps2 * (1.0-T.dot(p2));
// 	T -= p2 * (T.dot(p2)/p2.dot(p2));
// 	T.normalize();
// 	l = z*p2 + sqrt(1.0-z*z)*T;
      }
      else {

	//return(SbColor(1,0,0));

	// choose sample point using flat triangle

	// 1. find p1, p2, p3, points on unit sphere surrounding x
	p1 += r.x;
	p2 += r.x;
	p3 += r.x;

// 	cB = 1.0-sqrt(1.0-eps2);
// 	cC = (1.0-cB)*eps1;
// 	cA = 1.0-cB-cC;

// 	xPP = cA*p1 + cB*p2 + cC*p3;
    
// 	// 4. Find light direction, l
// 	l = xPP - r.x;

        double t, u, v;
        if (!intersect_ray_tri(r.pos, l, p1, p2, p3, &t, &u, &v))
          continue;
        
        lPP = t;
      }

      // 2d adjustment
      if (gv->useDimensions<3)
        l[1] = 0.0;
            
//       // dont illuminate lights below horizon
//       if (ng.dot(l) <= 0.0){
// 	continue;
//       }

//       // don't illuminate area that think they are facing away
//       dot1 = n.dot(l);
//       if (dot1<=0.0)
// 	continue;      

//       // trace from point to light point
//       // i add a small amount to make it not intersect overlapping surfaces
//       const float t_eps = 0.002;
//       Ray lr = Ray(r.x, l, r.ctx, r.prim);
      
//       // get distance to triangle
//       if (!lTri->intersectRaySimple(lr, t_eps, 1e8))
// 	continue; // occluded

      // NOTE: we must intersect the light to get lr.t,
      // (for APPROX_SAMPLING and ray marching)
      
//       double value=1;   // 1/p(w)
//       double pdf;
      
      if (sampleMethod[lNum] == ACCURATE_SAMPLING){
	pdfLight = 1/cv[lNum].sarea;
      }
      else {

	// calculate edges and normal
	SbVec3f edge1 = p2-p1;
	SbVec3f edge2 = p3-p1;
	SbVec3f nPP = edge1.cross(edge2);

	double area = 0.5*nPP.normalize();

// 	nP = lTri->getNormalPre();
// 	dot2 = -nP.dot(l);

        // light may point orthogonal to point
//         if (dot2 <= 0)
//           continue;
	
	double dot3 = -nPP.dot(l);
	if (dot3<=0.0)
	  continue;
    
	// 6. Find p(x) =                  ||x-x"||^2 * cosO'
	//                 -----------------------------------
	//                  area(p1p2p3) * ||x-x'||^2 * cosO"

	lP = lr.t;
	double prob = (gv->meters2(lPP*lPP) * dot2) /
	  (gv->meters2(area) * gv->meters2(lP*lP) * dot3);

        pdfLight = (gv->meters2(lP*lP) * prob)/dot2;
      }


        brdfCosT = brdfColor*brdfVal*dot1;
        newPdf = lightWeight*pdfLight + brdfWeight*pdfBrdf;
      }
      
      // phone emission from RRT pg. 124
      static float lightExp = getenv("PANE_LIGHT_EXP") ? atof(getenv("PANE_LIGHT_EXP")) : 1;
      
      SbColor Le = lTri->getEmission() / // in W
	gv->meters2( lTri->area ) * M_1_PI * (lightExp+1)/2*powf(dot2,lightExp-1);
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light

      if (newPdf<1e-6)
        continue;

      // weight light
      value /= newPdf;

      // ray marching absorbtion and attenuation
      SbVec3f lx = r.x + lr.t*l;
      gv->pane->applyAttenuation(r, r.x, lx, lr.t, Le);
      
      // debugging
      if (!finite(value) || !finite(brdfCosT)){
	cerr << "Problem in areaLightImpContrib: non-finite value" << endl;
	value = 0.0;
      }
      
      // compute color
      // shirley p.120 eq. 15.5
      color += vecScale(brdfCosT*value, Le);
      
      if (DEBUG_STATE){
	cerr << flush;
	print(" value: %f ", value);
	printCol(color);
	cerr << endl;
      }
      
    } // end stratified sampling x
  } // end stratified sampling x

  color *= invSqrtSamples*invSqrtSamples;
  
  return color;
}









// Calculates area lights contribution for participating medium
// Uses multiple luminaire importance sampling
// Discussed in Peter Shirleys Realistic Ray Tracing 1st ed. p. 124
// Basically, we simulate a discrete random variable
// to sample a probability mass function, where the masses
// are proportional to the importance (proximity and direction)
// of the lights.
//
SbColor Pane::volAreaImpLightContrib(Ray &r, int samples){

//   if (gv->useDimensions==2)
//     return(areaLightContrib2D(r, n, ng));

  SbColor color = SbColor(0.0, 0.0, 0.0);
  SbVec3f l, xP, xPP, nP;
  EmissiveTriangle *lTri;

  int numLights = litTris.size();
  if (numLights == 0 || gv->lightSamples <= 0 || samples==0)
    return(color); // black

  double cA, cB, cC;
  double e_i, dot1, dot2, dot3;
  double lP, lP2, lPP;

  float *weights = (float *)alloca(sizeof(float)*numLights);
  
  /////////////////////////
  // compute weights[]
  /////////////////////////
  double eSum = 0;
  int lNum = 0;
  for (int ai=0; ai<litTris.size(); ai++,lNum++){
    
    // L*cosO*cosO'*A/dist^2
    // compute Lavg
    lTri = &litTris[ai];
    e_i = lTri->getEmisAvg()/lTri->area;  // ~ L(i)

    // Figure purposes - only allow certain lights
    if (gv->figureMode && gv->useGlassModel==0 &&
	ABSF(e_i-0.15)>0.01)
      continue;
    
    // find normal
    nP = lTri->getNormalPre();

    //                      /\
    //                     /  \
    //                    /    \
    // 	     nP<------xP \
    //                  '---/----'
    // 		    /
    //           nPP      /
    //            ^\  p3 /l
    //              \ /\/        
    //               \ /\
    //              / xPP\
    //           p2'-/----p1
    //              x

    // find center
    xP = (lTri->vert(X)+lTri->vert(Y)+lTri->vert(Z)) * (1./3.);

    // make rough estimate
    l = xP-r.x;
    lP = l.normalize();

    // calculate angle of light normal and light vector
    dot2 = -nP.dot(l);
    
    // check to see if luminaire pointing away
    if (dot2 <= 0.0){
      weights[lNum] = 0.0;
      continue;
    }

    // calculate projected solid angle (max pi, area of unit circle)
    float w1 = lTri->area / (lP*lP); // max projected area
    float w2 = dot2;          // approximate projected area scalar components
    e_i *= w1 * w2;

    weights[lNum] = e_i;
    eSum += e_i;
  }

  if (eSum == 0.0)
    return(color); // black
  
  for (lNum=0; lNum<numLights; lNum++){
    weights[lNum] /= eSum;
  }

  /////////////////////////////////
  // importance sample the lights
  /////////////////////////////////

  // stratified sampling
//   int sqrt_samples = (int)sqrt((float)gv->lightSamples);
  int sqrt_samples = (int)sqrt((float)samples);
  
  double invSqrtSamples = 1./sqrt_samples;
  
//   // use one sample in final gather
//   if (simple)
//     sqrt_samples = 1;

  for (int sampx=0; sampx<sqrt_samples; sampx++){
    for (int sampy=0; sampy<sqrt_samples; sampy++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) * invSqrtSamples;
      double eps2 = (sampy + r.ctx->mydrand48()) * invSqrtSamples;
      //double eps1 = drand48(), eps2 = drand48();

      // pick light
      // could be faster with bisectional search instead of linear
      eSum = 0.0;
      double eSumNext = 0.0;
      int lastGood = 0;
      for (lNum=0; lNum<numLights-1; lNum++){
	if (weights[lNum]<=0) continue;
	lastGood = lNum;
	eSumNext += weights[lNum];
	if (eps1<=eSumNext) break;
	eSum = eSumNext;
      }
      if (weights[lNum]<=0.0) lNum = lastGood;

      eps1 = (eps1 - eSum) / weights[lNum];

      lTri = &litTris[lNum];

      // Generate sample on solid angle of triangle
      // from "Stratified Sampling of Spherical Triangles"
      // by James Arvo
      // Code mostly glibbed from Java version

      // this would be better if it picked
      // a point on the CLIPPED luminaire
      // oh well, future work
           
      // 1. find p1, p2, p3, points on unit sphere surrounding x
      SbVec3f p1, p2, p3;
      p1 = lTri->vert(0)-r.x;
      p1.normalize();
      p2 = lTri->vert(1)-r.x;
      p2.normalize();
      p3 = lTri->vert(2)-r.x;
      p3.normalize();

      // choose sample point using flat triangle

      // 1. find p1, p2, p3, points on unit sphere surrounding x
      p1 += r.x;
      p2 += r.x;
      p3 += r.x;
      
      cB = 1.0-sqrt(1.0-eps2);
      cC = (1.0-cB)*eps1;
      cA = 1.0-cB-cC;
      
      xPP = cA*p1 + cB*p2 + cC*p3;
	
      // 4. Find light direction, l
      l = xPP - r.x;
      
      // 2d adjustment
//       if (gv->useDimensions<3)
// 	l[1] = 0.0;
    
      lPP = l.normalize();

      // trace from point to light point
      // i add a small amount to make it not intersect overlapping surfaces
      const float t_eps = 0.002;
      Ray lr = Ray(r.x, l, r.ctx);
      if (!scenePrim->intersectRaySimple(lr, t_eps, 1e8) || lr.prim != lTri)
	continue; // occluded

      // NOTE: we must intersect the light to get lr.t,
      // (for APPROX_SAMPLING and ray marching)

      double value;
      
      // calculate edges and normal
      SbVec3f edge1 = p2-p1;
      SbVec3f edge2 = p3-p1;
      SbVec3f nPP = edge1.cross(edge2);
      
      double area = 0.5*nPP.normalize();
      
      nP = lTri->getNormalPre();
      dot2 = -nP.dot(l);
      
      // FIXME: test seems needless
      double dot3 = -nPP.dot(l);
      if (dot3<0.0)
	dot3 = -dot3;
      
      // 6. Find p(x) =                  ||x-x"||^2 * cosO'
      //                 -----------------------------------
      //                  area(p1p2p3) * ||x-x'||^2 * cosO"
      
      lP = lr.t;	
      double prob = (gv->meters2(lPP*lPP) * dot2) /
	(gv->meters2(area) * gv->meters2(lP*lP) * dot3);
      
      value = dot2 / gv->meters2(lP*lP) / prob;
      // FIXME: this equation can be drastically simplified

      SbColor Le = lTri->getEmission() / // in W
	gv->meters2( lTri->area ) * M_1_PI;
      // Le = E(x') / pi = Phi_l/area / pi       note: Phi_l = power of light

      // weight light
      value /= weights[lNum];

      // ray marching absorbtion and attenuation
      SbVec3f lx = r.x + l*lr.t;
      gv->pane->applyAttenuation(r, r.x, lx, lr.t, Le);
      
      // debugging
      if (!finite(value))
	cerr << "Problem in areaLightImpContrib: non-finite value\n";
      
      // compute color
      // shirley p.120 eq. 15.5
      color += Le*value;
      
      if (DEBUG_STATE){
	cerr << flush;
	print(" value: %f ", value);
	printCol(color);
	cerr << endl;
      }
      
    } // end stratified sampling x
  } // end stratified sampling x

  color /= sqrt_samples*sqrt_samples;
  
  return(color);
}


SbColor Pane::sphereLightContrib(Ray &r, SbVec3f &n, SbVec3f &ng, bool simple){
  
  SbColor slcolor = SbColor(0, 0, 0);
  float value;
  SbVec3f l;

  for (int pI=0; pI<litSpheres.size(); pI++){

    const Sphere &sph = spheres[litSpheres[pI]];
    SbVec3f w = sph.pos()-r.x;
    float w2 = w.dot(w);
    float rad2 = sph.rad()*sph.rad();
    if (w2<rad2) continue;
    ONB onb(w);
    double cos_a_max = sqrt(1 - rad2/w2);
    double frac = (1-cos_a_max);
    double omega = 2*M_PI*frac;  // exact solid angle, computed analytically
    float pdfLight = 1/omega;
    
    // stratified sampling
    int lightSamples = max2((int)(gv->lightSamples*min2(frac*50,1.)), 1);
    int sqrtSamples = simple ? 1 : (int)ceil(sqrtf(lightSamples));
    double invSqrtSamples = 1./sqrtSamples;
    SbColor color(0,0,0);

    SbColor Le = sph.getMat()->getEmission() /   // W
      gv->meters2((4*M_PI)*rad2) * M_1_PI;
    
    for (int sampx=0; sampx<sqrtSamples; sampx++){
      for (int sampy=0; sampy<sqrtSamples; sampy++){

        // jittered
        double eps1 = (sampx + r.ctx->mydrand48()) * invSqrtSamples;
        double eps2 = (sampy + r.ctx->mydrand48()) * invSqrtSamples;
        
        Material *mtl = r.prim->getMat(r);
        BRDF *brdf = mtl->getBRDF();
        SbVec3f brdfCosT;
        float newPdf;

        // Multiple Importance Sampling (MIS)
        float lightWeight = mtl->getGlossiness() ? 0.5 : 1;
        float brdfWeight = 1 - lightWeight;
        if (1<=lightWeight || r.ctx->mydrand48()<lightWeight){

          // sample light
          
          // compute light direction, l
          double cos_a = 1-eps1+eps1*cos_a_max;
          double sin_a = sqrt(1-cos_a*cos_a);
          double phi = (2*M_PI)*eps2;
          SbVec3f hemiDir(cos(phi)*sin_a,
                          sin(phi)*sin_a,
                          cos_a);
          l = onb.toWorldSpace(hemiDir);
          l.normalize();

          // dont use lights below horizon
          if (ng.dot(l) <= 0.0){
            continue;
          }

          // don't illuminate areas that think they are facing away
          float dot1 = n.dot(l);
          if (dot1<=0.0){
            continue;
          }

          // trace shadow ray
          Ray sr(r.x, l, r.ctx);
          const float t_eps = 0.002;
          //sr.ignore(r.prim);
          if (!scenePrim->intersectRaySimple(sr, t_eps) || sr.prim != &sph){
            continue;
          }
          
          float pdfBrdf;
          brdf->GetBRDF(r.dir, l, n, ng, r, brdfCosT, pdfBrdf);
          newPdf = lightWeight*pdfLight + brdfWeight*pdfBrdf;
          
        } else {

          // use brdf sampling
          
          double brdfVal, pdfBrdf, brdfOverPdfVal, brdfOverPdfCosTval;
          SbVec3f brdfColor;
          double eps3 = r.ctx->mydrand48();
          brdf->GetSampleDir(eps1, eps2, eps3, r.dir, n, ng, r, l,
                             brdfVal, pdfBrdf, brdfOverPdfVal, brdfOverPdfCosTval,
                             brdfColor);

          if (pdfBrdf<=0){
            continue;
          }

          // dont use lights below horizon
          if (ng.dot(l) <= 0.0){
            continue;
          }

          // don't illuminate areas that think they are facing away
          float dot1 = n.dot(l);
          if (dot1<=0.0){
            continue;
          }

          // trace shadow ray
          Ray sr(r.x, l, r.ctx);
          const float t_eps = 0.002;
          //sr.ignore(r.prim);        
          if (!scenePrim->intersectRaySimple(sr, t_eps) ||
              !sr.prim->isOfType(Primitive::PRIMTYPE_SPHERE)){
            continue;
          }

          // recompute light pdf (may be a different emissive sphere)
          const Sphere &sph = *(Sphere *)sr.prim;
          if (!sph.getMat()->emissive) continue;
          SbVec3f w = sph.pos()-r.x;
          float w2 = w.dot(w);
          float rad2 = sph.rad()*sph.rad();
          if (w2<rad2) continue;
          double cos_a_max = sqrt(1 - rad2/w2);
          double frac = (1-cos_a_max);
          double omega = 2*M_PI*frac;  // exact solid angle, computed analytically
          float pdfLight = 1/omega;

          SbColor Le = sph.getMat()->getEmission() /   // W
            gv->meters2((4*M_PI)*rad2) * M_1_PI;

          // divide by numSpheres to counter that we are adding
          // total contribution numSpheres times
          brdfCosT = brdfColor*brdfVal*dot1 * (1/litSpheres.size());
          newPdf = lightWeight*pdfLight + brdfWeight*pdfBrdf;

        }

        if (newPdf<1e-6)
          continue;
        
        value = 1/newPdf;
        color += vecScale(brdfCosT*value, Le);
      }
    }

    slcolor += color*(invSqrtSamples*invSqrtSamples);
  }

  return slcolor;
}
