#include <stdlib.h>
#include "Pane.h"
#include "debug.h"
#include "FileUtils.h"
#include "FontWriter.h"
#include "rgbe.h"

using namespace std;

//'''''''''''''''''WW'WW''W'WWWW''WWWW'
//                 WW WW  W WW  W WW  W
//W WW  WW   WWW  WWW WW  W WW  W WW  W
//WW   WWWW W WW W WW WWWWW WW  W WWWW 
//WW   WW   W WW W WW WW  W WW  W WW  W
//WW    WWW  WWW  WWW WW  W WWWW  WW  W
//                                     
//'''''''''''''''''''''''''''''''''''''

void Pane::readHDR(){

  cerr << "Reading " << gv->inImgFile << " into framebuffer.\n";
  
  // read file
  FILE *f = fopen(gv->inImgFile.c_str(), "rb");
  if (!f){
    cerr << "Error opening " << gv->inImgFile << ": " << strerror(errno) << endl;
    return;
  }
    
  RGBE_ReadHeader(f, &gv->imgTotalDim[X], &gv->imgTotalDim[Y], NULL);
  float *image = (float *)malloc(sizeof(float)*3*gv->imgTotalDim[X]*gv->imgTotalDim[Y]);
  RGBE_ReadPixels_RLE(f, image, gv->imgTotalDim[X], gv->imgTotalDim[Y]);

  // read comment
  char *line = (char *)malloc(sizeof(char)*1024);
  size_t n;
  inputFileComment.clear();
  if (getline(&line, &n, f)<=0) // read new line
    cerr << "Error reading " << gv->inImgFile << endl;
  while (getline(&line, &n, f)>=0){
    string s = line;
    if (!s.empty())
      inputFileComment += line;
  }
  free(line);
  
  fclose(f);

  // allocate new pix
  delete [] pix;
  delete [] pixDepth;
  delete [] pixWeight;
  int size = gv->imgTotalDim[X]*gv->imgTotalDim[Y];
  pix       = new SbColor[size]; // should i initialize this?
  pixDepth  = new float[size];
  pixWeight = new float[size];

  gv->imgDim[X] = gv->imgTotalDim[X]/gv->subPixDim;
  gv->imgDim[Y] = gv->imgTotalDim[Y]/gv->subPixDim;
  
  // copy into frame buffer
  int k=0;
  for (int j=gv->imgTotalDim[Y]-1; j>=0; j--)
    for (int i=0; i<gv->imgTotalDim[X]; i++){
      SbColor c(&image[(k++)*3]);
      setSubPix(i, j, c, 0.1);
    }

  delete image;
}

// read filter image
int Pane::readGlareFilter(const string &fn, int *fdim){

  //char fn[] = "filter.hdr";

  cerr << "Reading " << fn << " into memory.\n";
  
  // read file
  FILE *f = fopen(fn.c_str(), "rb");
  if (!f){
    cerr << "Error reading " << fn << ": " << strerror(errno) << endl;
    return(0);
  }
    
  RGBE_ReadHeader(f, &(fdim[X]), &(fdim[Y]), NULL);
  float *image = (float *)malloc(sizeof(float)*3*fdim[X]*fdim[Y]);
  RGBE_ReadPixels_RLE(f, image, fdim[X], fdim[Y]);
  fclose(f);

  // allocate new pix
  filtpix = new SbColor[fdim[X]*fdim[Y]]; // should i initialize this?
  
  // copy into frame buffer
  int k=0;
  for (int j=fdim[Y]-1; j>=0; j--)
    for (int i=0; i<fdim[X]; i++)
      filtpix[j*fdim[X]+i].setValue(&image[(k++)*3]);

  return(1);
}

//'''''''''''''''''''''WW''''''''WWW'''''''''''''''''''''''''''''''''''WW'
//                     WW       WW  W                                  WW 
// WWW W WW  WW   WWW WWWW  WW  WW     WW   WWWW W  WWWW W   WW  WWW  WWWW
//WW W WW   WWWW W WW  WW  WWWW WW    WW W  WW WW W WW WW W WWWW WW W  WW 
//WW   WW   WW   W WW  WW  WW   WW  W WW W  WW WW W WW WW W WW   WW W  WW 
// WWW WW    WWW  WWW  WW   WWW  WWW   WW   WW WW W WW WW W  WWW WW W  WW 
//                                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

string Pane::createComment(){

  // get hostname
  char machinename[100];
  gethostname(machinename, 100);

  // create comment
  ostringstream comment;
  
  int totalRenderTime = gv->times.octreeCreation + gv->times.photonShooting +
    gv->times.render + gv->times.photonBalancing + gv->times.photonIrradiance;
  float renderCPUMs = gv->times.renderTicks / sysconf(_SC_CLK_TCK) * 1000.0;
  
  gettimeofday(&gv->times.progEnd, NULL);   // end delay measure
  int totalWallTimeMs = delay(gv->times.progStart, gv->times.progEnd);
  gv->times.wallTime = totalWallTimeMs;
  
  comment << "# " << getNameVersionDate() << "\n";
  comment << "# by " << getAuthorName() << "\n";
  comment << "# " << getAuthorUrl() << "\n";
  comment << "# Date: " << getCurrentTime();
  comment << "#\n";
  comment << "# " << gv->numSamples << " samples, "
	  << gv->lightSamples << " light samples, "
	  << gv->numThreads << " threads\n";
  comment << "# Octree time: "      << gv->times.octreeCreation << "ms "
	  << " Photon shooting: "   << gv->times.photonShooting << "ms "
	  << " Photon balancing: "  << gv->times.photonBalancing << "ms "
	  << " Photon irradiance: " << gv->times.photonIrradiance << "ms "
	  << " Rendering: "         << gv->times.render << "ms\n";
  comment << "# Rendering CPU time (CPU time): " << renderCPUMs << "ms / ";
  comment << renderCPUMs/1000.0 << "s / ";
  comment << renderCPUMs/1000.0/60.0 << "m / ";
  comment << renderCPUMs/1000.0/60.0/60.0 << "h, ";
  comment << "Ticks: " << gv->times.renderTicks << " ";
  comment << "Ticks/s: " << sysconf(_SC_CLK_TCK) << "\n";
  comment << "# Total render time (wall, last pass): " << totalRenderTime << "ms / ";
  comment << totalRenderTime/1000.0 << "s / ";
  comment << totalRenderTime/1000.0/60.0 << "m / ";
  comment << totalRenderTime/1000.0/60.0/60.0 << "h\n";
  comment << "# Cumulative time (wall, all passes) : " << totalWallTimeMs << "ms / ";
  comment << totalWallTimeMs/1000.0 << "s / ";
  comment << totalWallTimeMs/1000.0/60.0 << "m / ";
  comment << totalWallTimeMs/1000.0/60.0/60.0 << "h\n";
  comment << "# on " << machinename << "\n";
  if (gv->toneMapping){
    comment << "# Yi = " << lastImageYi
	<< " (Log average luminance used for tone mapping)\n";
  }
  else
    comment << "# (No tone mapping, though Yi = " << lastImageYi << ")\n";
  comment << "# Average luminance = " << lastImageAvgYi	<< "\n";

  comment << "# Comment: " << gv->comment << "\n";
  comment << "# ";
  
  // insert environment variables
  cerr << "Environment:\n";
  bool anyEnviron = false;
  for (int i=0; environ[i]; i++)
    if (!strncmp(environ[i], "PANE", 4)){
      if (!anyEnviron){
        comment << "env ";
        anyEnviron = true;
      }
      comment << environ[i] << " ";
    }
  
  comment << gv->cmdLine << "\n";

  // tack on camera info
  if (gv->inImgFile.empty()){
  
    comment << "# Camera:\n";
    extern string getCameraString(GlobalVars *gv);
    string camInfo = getCameraString(gv);
    comment << camInfo;
  }
  
  // tack on input file comment
  if (!inputFileComment.empty())
    comment << "#\n#\n######################################################################\n"
            << "# Input file comment:\n#\n" << inputFileComment;

  return comment.str();
}

//'''''''''''''WW''WW'''''''WWWW'''''''''''''''WW'WW'
//                 WW        WW               WW   WW
//WW    W W WW WW WWWW  WW   WW  WWWW W   WWW WW   WW
//WW WW W WW   WW  WW  WWWW  WW  WW WW W W WW WW   WW
//WW WW W WW   WW  WW  WW    WW  WW WW W W WW WW   WW
// WW WW  WW   WW  WW   WWW WWWW WW WW W  WWW  WW WW 
//                                         WW        
//'''''''''''''''''''''''''''''''''''''''WWW'''''''''
//
// Writes image to disk in PNG format.
//
void Pane::writeImg(){

  if (currentlyRendering ||
      currentlyShooting ||
      currentlyRenderingPPMRange) {
    writeWhenDone = true;
    cerr << "\nWill write image when finished rendering.\n";
    return;
  }

  // if configured, output iv file instead
  if (gv->renderIv){
    writeIv();
    return;
  }

  cerr << "\nWriting " << gv->imgFile << " ...";
  ofstream out;

  // for matting
  if (gv->makeMatte){
    for (int i=0; i<gv->imgTotalDim[0]*gv->imgTotalDim[1]; i++)
      if (pixDepth[i]<0.0){
	pix[i] = gv->worldBG;
	pixWeight[i] = 1.0;
      }
  }

  // write hdr
  string comment;
  if (gv->inImgFile.empty()){
    
    // create comment
    // (will not include accurate tone mapping info but better than nothing)
    comment = createComment();
    
    FileUtils::writeHDR(gv->imgTotalDim, gv->imgFile,
                        pix, gv->createBackup, comment.c_str());
  }

  // write depth
  if (gv->depthImage)
    FileUtils::writeDepth(gv->imgTotalDim, gv->imgFile, pixDepth,
                          gv->createBackup);
  
  // back up old file if it exists
  string oldFile = FileUtils::checkForOverwrite(gv->imgFile, gv->createBackup);

  // tone map the image
  if (gv->debugPixel[X]>=0 && gv->debugPixel[Y]>=0){
    if (gv->debugPixel[X]>=gv->imgDim[X] ||
        gv->debugPixel[Y]>=gv->imgDim[Y]){
      cerr << "Debug pixel ("
           << gv->debugPixel[X] << ", "
           << gv->debugPixel[Y] << ") out of image. Not outputting.\n";
    } else {
      cerr << "Before tonemapping, pix("
           << gv->debugPixel[X] << ","
           << gv->debugPixel[Y] << ")="
           << getPix(gv->debugPixel[X], gv->debugPixel[Y]) << "\n";
    }
  }

  SbColor *pixOut = NULL; // pointer for reduced resolution output
  
  if (toneMapped == 0)
    toneMap(pixOut);

  if (!pixOut)
    downSampleImg(pix, gv->imgTotalDim[0], gv->imgTotalDim[1],
                  pixOut, gv->imgDim[0], gv->imgDim[1], gv->subPixDim);
  
  if (gv->debugPixel[X]>=0 && gv->debugPixel[Y]>=0){
    int fixY = gv->imgDim[Y]-gv->debugPixel[Y]-1;
    int idx = gv->imgDim[X]*fixY+gv->debugPixel[X];
    cerr << "After tonemapping, pix("
	 << gv->debugPixel[X] << ","
	 << gv->debugPixel[Y] << ")="
	 << pixOut[idx] << "\n";
  }
  
  // recreate comment
  comment = createComment();

  // add watermark (disable with -w'')
  FontWriter::write(gv->imgDim, pixOut, gv->waterMark);
  
  bool res = FileUtils::writePNG(gv->imgDim, gv->imgFile,
                                 pixOut, gv->gamma, comment.c_str());
  delete [] pixOut;

  if (!res){
    cerr << "Error writing output: " << gv->imgFile << " !" << endl;
    return;
  }
  
  cerr << "done.\n";
    
  if (!gv->renderQuit){
    
    // choose appropriate viewer
    string f;

    // parse PATH
    string path = getenv("PATH");
    string delim = ":";
    vector<string> paths;
    string::size_type lastPos = path.find_first_not_of(delim, 0);
    string::size_type pos = path.find_first_of(delim, lastPos);
    while (lastPos != string::npos || pos != string::npos){
      paths.push_back(path.substr(lastPos, pos-lastPos));
      lastPos = path.find_first_not_of(delim, pos);
      pos = path.find_first_of(delim, lastPos);
    }

    // search paths
    vector<string> progs;
    progs.push_back("xv");
    progs.push_back("display");
    for (int i=0; i<progs.size() && !f.size(); i++)
      for (int j=0; j<paths.size(); j++)
        if (!access((paths[j]+"/"+progs[i]).c_str(), R_OK|X_OK)){
          f = progs[i];
          break;
        }

    // check if suitable path was found
    if (!f.size()) {
      cerr << "No appropriate viewer found!\n"
	   << "Try installing imagemagick or putting xv in your path.\n";
      return;
    }

    f = f + " " + gv->imgFile + " " + oldFile;

    // add reference image if it exists
    string ref = "ref.png";
    struct stat file_stat;
    if (stat(ref.c_str(), &file_stat) == 0)
      f += " " + ref;
        
    if (system(f.c_str())<0)
      cerr << "Error executing command: " << f << endl;
  }
}

SbColor* Pane::downSampleImg(SbColor *in, int inX, int inY,
                             SbColor *&pixOut, int outX, int outY, int subPixDim) {

  // outX = inX/subPixDim
  // outY = inY/subPixDim
  
  // down sample the image
  if (!pixOut) pixOut = new SbColor[outX*outY];

  static bool subpixTent = getenv("PANE_SUBPIX_TENT");

  if (!subpixTent){
    
    // simple box filter
    float invNumPix = 1.f/(subPixDim*subPixDim);
    for (int i=0; i<outX; i++){
      for (int j=0; j<outY; j++){
        SbColor avg(0,0,0);
        for (int sx=0; sx<subPixDim; sx++){
          for (int sy=0; sy<subPixDim; sy++){
            int si = i*subPixDim + sx;
            int sj = j*subPixDim + sy;
            avg += pix[sj*inX + si];
          }
        }
        avg *= invNumPix;
        pixOut[j*outX + i] = avg;
      }
    }

  } else {
    
    // better filter (tent)
    // slightly softer/blurrier
    float invSubPixDim = 1.f/subPixDim;
    for (int i=0; i<outX; i++){
      for (int j=0; j<outY; j++){

        int si0 = (int)((i+0.5-1.001)*subPixDim);
        int sj0 = (int)((j+0.5-1.001)*subPixDim);
        int si1 = (int)((i+0.5+1.001)*subPixDim)+1;
        int sj1 = (int)((j+0.5+1.001)*subPixDim)+1;
        si0 = max2(si0, 0);
        sj0 = max2(sj0, 0);
        si1 = min2(si1, inX);
        sj1 = min2(sj1, inY);
      
        float weightSum=0;
        SbColor avg(0,0,0);

        for (int si=si0; si<si1; si++) {
          for (int sj=sj0; sj<sj1; sj++) {

            SbVec2f pixCrd(i+0.5, j+0.5);
            SbVec2f subPixCrd(si+0.5, sj+0.5);
            subPixCrd *= invSubPixDim;

            // this could be faster if I pulled out one of the filter weights out
            // of loop...
            // separated tent filter
            SbVec2f dist = pixCrd-subPixCrd;
            dist[0] = fabs(dist[0]);
            dist[1] = fabs(dist[1]);
            float weight = tentFilter(dist[0]) * tentFilter(dist[1]);

            // mitchell filter (be sure to radius *=2 above!)
            //float distn = dist.length();
            //float weight = mitchellFilter(distn);
           
            avg += weight*pix[sj*inX + si];
            weightSum += weight;
          }
        }
          
        if (weightSum)
          avg /= weightSum;
        pixOut[j*outX + i] = avg;
      }
    }

  }

  return pixOut;
}
