#include "Pane.h"
#include "debug.h"
#include "FileUtils.h"
#include <alloca.h>
#include <stdlib.h>

using namespace std;

SoMaterial   *currMaterial;
SoGroup      *currGroup;
SoShapeHints *currShapeHints;

// Material CB Handler
// Stores pointer to most recent SoMaterial node in
// scene graph traversal.
SoCallbackAction::Response
materialCBHandler(void *data, SoCallbackAction *action,
                  const SoNode *node){
  currMaterial = (SoMaterial *)node;
  
  return(SoCallbackAction::CONTINUE);
}

// Group CB Handler
// Stores pointer to most recent SoGroup node in
// sene graph traversal.
SoCallbackAction::Response
groupCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  currGroup = (SoGroup *)node;
  
  return(SoCallbackAction::CONTINUE);
}

// ShapeHints CB Handler
// Stores pointer to most recent SoShapeHints node in
// sene graph traversal.
SoCallbackAction::Response
shapeHintsCBHandler(void *data, SoCallbackAction *action,
		   const SoNode *node){
  currShapeHints = (SoShapeHints *)node;
  
  return(SoCallbackAction::CONTINUE);
}

//'''''''''''''''''WW'''''''''''WWWW'''''''WW'WW'
//                 WW            WW       WW   WW
//W WW  WW  WWW   WWW  WW  W WW  WW  WW W WW   WW
//WW   WWWW WW W W WW WWWW WW    WW  WW W WW   WW
//WW   WW   WW W W WW WW   WW    WW  WW W WW   WW
//WW    WWW WW W  WWW  WWW WW   WWWW  WW   WW WW 
//                                               
//'''''''''''''''''''''''''''''''''''''''''''''''
// (renderIv)
//
// Like render(), but instead of rendering a 2D image
// renders a 3D Inventor scene,
// computing radiance values for each vertex
// in the scene.
// (It actally computes a radiance value for each face
// and distributes the color to the vertices.)
// The scene's color values is edited in place,
// inserting SoMaterial nodes when appropriate.
// 
// Unlike render(), renderIv() performs tone mapping
// before returning.
//
// This procedure is NOT multi-threaded.
//

void Pane::renderIv(thread_context *ctx){

  // tone mapping variables
  const double k = 683.0;  // k constant (cd/W) (Peter Shirley)
  double V;
  const double Yw = 80; // maximum display luminance (cd/m^2) (PS)
  double Yi = 0.0;         // world adaptation luminance
  double avgYi = 0.0;      // world average luminance
  double cX, cY, cZ;       // tristimulous values
  double x, y, W, s;       // misc
  const double l[3] = { 620.0, 545.0, 465.0 }; // lambda (nm)
  double r, g, b;
  int numpix = 0;

  const double xw=0.3333, yw=0.3333;
  const double d = (0.4905*0.65 + 0.9928*0.94 + 0.0770*1.0) /
                   (0.2126+0.7152 +0.0722);  // 1.3291
  const double e = d;
  const double f = d;
//   const double xb[3] = { 0.4124*d, 0.3576*e, 0.1805*f }; // xbar
//   const double yb[3] = { 0.2126*d, 0.7152*e, 0.0722*f }; // ybar
//   const double zb[3] = { 0.0193*d, 0.1192*e, 0.9505*f }; // zbar
  const double xb[3] = { 0.4124, 0.3576, 0.1805 }; // xbar
  const double yb[3] = { 0.2126, 0.7151, 0.0721 }; // ybar
  const double zb[3] = { 0.0193, 0.1192, 0.9505 }; // zbar
  const double vb[3] = { 0.01477, 0.497, 0.631 }; // vbar
  const double dlambda[3] = { 1.0, 1.0, 1.0 };

  // feedback
  struct timeval start, end;
  cerr << "Rendering...\n";
  gettimeofday(&start, NULL); // start delay measure
  
  // loop over surfaces
  int totalTris = 0;
  int totalVerts = 0;
  int j;

  // Search for indexed face sets
  SoCallbackAction callback;
  currMaterial = NULL;
  callback.addPreCallback(SoMaterial::getClassTypeId(), materialCBHandler, this);
  callback.addPreCallback(SoShapeHints::getClassTypeId(), shapeHintsCBHandler,
			  this);

  SoSearchAction sa;
  sa.setType(SoIndexedShape::getClassTypeId(), TRUE);
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(gv->scene);
  SoPathList plist = sa.getPaths();
  int len = plist.getLength();
  cerr << len << " paths found\n";

  //W'''''''''''''''''''''''''''''''''''''''''''''''''''W'''''''
  //W                                                W  W       
  //W     W   W  WW      W  W W  WW   WW    WW   WW WWW WW   WWW
  //W    W W W W W W    W W W W Wmmm W      W W W W  W  W W  WW 
  //WWWW  W   W  WW      W   W   WW  W      WW   WW  W  W W WWW 
  //'''''''''''''W''''''''''''''''''''''''''W'''''''''''''''''''
  for (int i=0; i<len; i++){

    // insert lightmodel & material node
    // render one pixel for triangle
    // add color to vertex colors (3),
    //   add 1 to each vertex surface counter
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();

    cerr << "indexed shape ";
    
    SoVertexProperty *vprop;
    vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
    if (!vprop){
      cerr << "Vprop not found\n";
      continue;
    }

    // get group
    currGroup = (SoGroup *)path->getNodeFromTail(1);
    currMaterial = NULL;
    currShapeHints = NULL;
    callback.apply(path);

    // get shape hints
    int reverseOrdering = 0;
    if (currShapeHints){
      if (currShapeHints->vertexOrdering.getValue() == SoShapeHints::CLOCKWISE)
	reverseOrdering = 1;
    }
    
    // get transformation
    SoGetMatrixAction ma = SoGetMatrixAction(SbViewportRegion());
    ma.apply(path);
    SbMatrix matrix = ma.getMatrix();
    SbMatrix matInvTran = matrix.inverse().transpose();

    // fix orderedRGBA
    vprop->orderedRGBA.setNum(0);
    int numVerts = vprop->vertex.getNum();
    vprop->materialBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);
    vprop->normalBinding.setValue(SoVertexProperty::PER_VERTEX_INDEXED);

    // fix material binding
    // shape->materialIndex.set1Value(j, shape->coordIndex[j]);
    int numCoordIdxs = shape->coordIndex.getNum();
    cerr << numCoordIdxs << " coord indexes\n";
    for (j=0; j<numCoordIdxs; j++)
      shape->materialIndex.set1Value(j, shape->coordIndex[j]);

    // currGroup->findChild(shape) is flawed,
    // as it assumes the group contains the shape.
    // the group might not contain the shape,
    // merely come before the shape (such as a switch)
    // and be empty or otherwise not contain the shape.
    
    // insert lightmodel
    SoLightModel *lightModel = new SoLightModel();
    lightModel->model = SoLightModel::BASE_COLOR;
    currGroup->insertChild(lightModel, currGroup->findChild(shape));

    // insert material
    SoMaterial *mat = new SoMaterial();
    if (currMaterial)
      mat->transparency = currMaterial->transparency;
    mat->diffuseColor.setNum(numVerts);
    for (j=0; j<numVerts; j++)
      mat->diffuseColor.set1Value(j, SbColor(0.0, 0.0, 0.0));
    currGroup->insertChild(mat, currGroup->findChild(shape));

    // insert material binding
    SoMaterialBinding *matBind = new SoMaterialBinding();
    matBind->value.setValue(SoMaterialBinding::PER_VERTEX_INDEXED);
    currGroup->insertChild(matBind, currGroup->findChild(shape));

    // insert shininess stuff after last indexed shape
    if (gv->renderFlags==60 && i==(len-1)){


      // is this working right? - kevin 11/9/04

      // append light model
      SoLightModel *phongLightModel = new SoLightModel();
      phongLightModel->model = SoLightModel::PHONG;
      currGroup->insertChild(phongLightModel, currGroup->findChild(shape)+1);

      // insert material
      SoMaterial *shinyMat = new SoMaterial();

      // if .8 shiny, we make (1-.8) transparent and 100% shiny
      if (currMaterial)
	shinyMat->transparency.set1Value(0,
			MAX2(1.0 - currMaterial->shininess[0], 0.01));
      else
	shinyMat->transparency.setValue(0.5);
      for (j=0; j<numVerts; j++)
	shinyMat->diffuseColor.set1Value(j, SbColor(0.0, 0.0, 0.0));
      // shinyMat->diffuseColor.setValue(0.0, 0.0, 0.0);
      shinyMat->specularColor.setValue(1.0, 1.0, 1.0);
      shinyMat->shininess.setValue(1.0);
      currGroup->insertChild(shinyMat, currGroup->findChild(shape)+2);

      // re-add shape
      currGroup->insertChild(shape, currGroup->findChild(shape)+3);

      // create off switch
      SoSwitch *shinySwitch = new SoSwitch();
      shinySwitch->addChild(mat);
      shinySwitch->addChild(shape);
      
      // add switch (so blurring tool can find it, i.e. it is last)
      currGroup->insertChild(shinySwitch, currGroup->findChild(shape)+4);
    }

    // initialize flags
    int flags;
    if (gv->renderFlags){
      flags = gv->renderFlags |	gv->approximate * Pane::C_APPROX;
    }
    else {
      flags = Pane::C_NORMAL;
      if (gv->algorithm==1) flags |= Pane::C_MDIFFUSE;
      if (gv->approximate)  flags |= Pane::C_APPROX;
      if (gv->stepSize)     flags |= Pane::C_VOLDIR;
      if (gv->stepSize && gv->numVPhotons>0) flags |= Pane::C_VOLINDIR;
    }

    
    // keep record of colors stored per vertex
    float *okayNormals = new float[numVerts];
    for (j=0; j<numVerts; j++)
      okayNormals[j] = -1.0;


    //WW''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''WW''WW'
    //WW                                                                WW     
    //WW     WW    WW   WWW      WW   WW W  WW  W WW    WW W  WW  W WW WWWW WW 
    //WW    WW W  WW W  WW W    WW W  WW W WWWW WW      WW W WWWW WW    WW  WW 
    //WW    WW W  WW W  WW W    WW W  WW W WW   WW      WW W WW   WW    WW  WW 
    //WWWWW  WW    WW   WWW      WW    WW   WWW WW       WW   WWW WW    WW  WW 
    //                  WW                                                     
    //''''''''''''''''''WW'''''''''''''''''''''''''''''''''''''''''''''''''''''

    typedef struct renderVert_st {
      int numVerts;
      SoMaterial *mat;
      SoVertexProperty *vprop;
      SbMatrix *matrix;
      SbMatrix *matInvTran;
      float *okayNormals;
      int flags;
      SoIndexedShape *shape;
      int numCoordIdxs;
      int reverseOrdering;
    } renderVert_st_t;

    renderVert_st_t passingVals;
    passingVals.numVerts = numVerts;
    passingVals.mat = mat;
    passingVals.vprop = vprop;
    passingVals.matrix = &matrix;
    passingVals.matInvTran = &matInvTran;
    passingVals.okayNormals = okayNormals;
    passingVals.flags = flags;
    passingVals.shape = shape;
    passingVals.numCoordIdxs = numCoordIdxs;
    passingVals.reverseOrdering = reverseOrdering;

    // launch threads
    MemberFunction *memberFunc = (MemberFunction *)alloca(
      sizeof(MemberFunction)*gv->maxThreads); // used to pass object pointer
    thread_context *ctx = (thread_context *)alloca(
      sizeof(thread_context)*gv->maxThreads);   // random number generators

    // debugging
    int cp = irradCache.numPoints();
    
    // subscoping to eliminate i loop variable namespace conflict
    {
      // scramble random number generators
      for (int i=0; i<gv->numThreads; i++){
	ctx[i].mydrand48 = RNG(mydrand48.lrand48());
	ctx[i].id = i;
	ctx[i].rays = 0;
	ctx[i].ticks = 0;
	ctx[i].other = (void *) &passingVals;
      }

      if (gv->numThreads == 1){
	renderVerts(&ctx[0]);
      }
      else {

	// Launch Threads
	for (int i=0; i<gv->numThreads; i++){

	  // pack object pointer, function pointer, and argument into struct
	  memberFunc[i].object = this;
	  memberFunc[i].pmf = &Pane::renderVerts;
	  memberFunc[i].ctx = &ctx[i];

	  // call wrapper function with struct
	  pthread_create(&gv->threads[i], NULL, Pane::renderTileWrapper,
			 (void*)&memberFunc[i]);
	}
  
	// wait for threads to finish
	for (int i=0; i<gv->numThreads; i++){
	  pthread_join(gv->threads[i], NULL);
	  renderRaysFired += ctx[i].rays;
	  gv->times.renderTicks += ctx[i].ticks;
	  gv->times.totalTicks += ctx[i].ticks;
	}
      }

      cerr << "\n";

      // debugging
      if (STOP_RENDERING)
	return;
    }

    
    
    // debugging
    int cp2 = irradCache.numPoints();
    cerr << cp2 << " irrcache points (+"
	 << cp2-cp << ")\n";
    
    totalVerts += numVerts;

    // normalize colors and find Yi
    for (j=0; j<numVerts; j++){

      // skip bad vertices
      if (okayNormals[j]<0.0)
	continue;

      r = mat->diffuseColor[j][R];
      g = mat->diffuseColor[j][G];
      b = mat->diffuseColor[j][B];

      // make sure pixel colors are not inf or nan
      if ( !(finite(r) && finite(g) && finite(b)) ){
	cerr << "Vertex color at (" << i << ", " << j << ") is invalid: ";
	printCol(SbColor(r,g,b));
	cerr << endl;

	if (!finite(r))
	  r = 0.0;
	if (!finite(g))
	  g = 0.0;
	if (!finite(b))
	  b = 0.0;

	mat->diffuseColor.set1Value(j, SbColor(r,g,b));
      }

      cY = k * ( yb[R]*r*dlambda[R] + yb[G]*g*dlambda[G] +
		 yb[B]*b*dlambda[B] );

      // minimum value of cY taken from Tumblin Rushmeier '93
      if (cY>3.18e-7){
        //cY = 3.18e-7;
	Yi += log10(cY);
        avgYi += cY;
	numpix++;
      }
      
    }

    delete [] okayNormals;

  } // end IndexedShape loop
  cerr << totalTris << " triangles, "
       << totalVerts << " verts found\n";



  //WWWWW'''''''''''''''''W''''W''''''''
  //  W                   WW  WW        
  //  W    W  WW   WW     W WW W  WW WW 
  //  W   W W W W Wmmm    W    W W W W W
  //  W    W  W W  WW     W    W  WW WW 
  //'''''''''''''''''''''''''''''''''W''
  
  // Log average of Y idea of Jack Tumblin and Holly Rushmeier
  // finish calculating Yi
  if (numpix){
    Yi /= numpix;
    avgYi /= numpix;
  }
  //Yi += 0.84;           // from Tumblin Rushmeier '93
  Yi = pow(10.0, Yi);

  // debugging
  if (gv->luminanceLogMean>0.0)
    Yi = gv->luminanceLogMean;
  
  lastImageYi = Yi;
  lastImageAvgYi = avgYi;
  
  cerr << "Using Yi = " << Yi << endl;
  
  double sf = 1.0/Yw * pow( ( 1.219 + pow(Yw/2.0, 0.4) ) /
			    ( 1.219 + pow(Yi, 0.4) ), 2.5);
//  sf *= gv->exposure;

  
  // loop over paths again
  for (int i=0; i<len; i++){

    // insert lightmodel & material node
    // render one pixel for triangle
    // add color to vertex colors (3),
    //   add 1 to each vertex surface counter
    SoPath *path = (SoPath*) plist[i];
    SoIndexedShape *shape = (SoIndexedShape *)path->getTail();

    cerr << "indexed shape\n";
    
    SoVertexProperty *vprop;
    vprop = (SoVertexProperty *)shape->vertexProperty.getValue();
    if (!vprop){
      cerr << "Vprop not found\n";
      continue;
    }

    // get material
    callback.apply(path);
    if (!currMaterial){
      cerr << "material not found\n";
      continue;
    }
 
    // tone map vertex colors
    int numVerts = vprop->vertex.getNum();
    for (j=0; j<numVerts; j++){

      r = currMaterial->diffuseColor[j][R]*gv->exposure;
      g = currMaterial->diffuseColor[j][G]*gv->exposure;
      b = currMaterial->diffuseColor[j][B]*gv->exposure;

      // do tone mapping
      if (gv->toneMapping){

	cX = k * ( xb[R]*r*dlambda[R] + xb[G]*g*dlambda[G] +
		   xb[B]*b*dlambda[B] );
	cY = k * ( yb[R]*r*dlambda[R] + yb[G]*g*dlambda[G] +
		   yb[B]*b*dlambda[B] );
	cZ = k * ( zb[R]*r*dlambda[R] + zb[G]*g*dlambda[G] +
		   zb[B]*b*dlambda[B] );
      
	V = 1700.0 * ( vb[R]*r*dlambda[R] + vb[G]*g*dlambda[G] +
		       vb[B]*b*dlambda[B] );
      
	if (cY<3.18e-7)
	  cY = 3.18e-7;

	if (log10(cY) < -2.0)
	  s = 0.0;
	else if (log10(cY) < 0.6){
	  double f = (log10(cY)+2.0)/2.6;
	  s = 3.0*f*f -  2.0*f*f*f;
	}
	else
	  s = 1.0;

	W = cY + cX + cZ;
	x = cX/W;
	y = cY/W;

	x = (1.0-s)*xw + s*(x+xw-0.3333);
	y = (1.0-s)*yw + s*(y+yw-0.3333);
	cY = 0.4468*(1.0-s)*V + s*cY;
	
	// Scale luminance
	cY *= sf;
	cX = x*cY/y;
	cZ = (1.0-x-y)*cY/y;

	r =  3.2406*cX +  -1.5372*cY +  -0.4986*cZ;
	g = -0.9689*cX +   1.8758*cY +   0.0415*cZ;
	b =  0.0557*cX +  -0.2040*cY +   1.0570*cZ;
	
      }
      
      // gamma correct
      r = clamp(r, 0.0, 1.0);
      g = clamp(g, 0.0, 1.0);
      b = clamp(b, 0.0, 1.0);

      if (fabs(gv->gamma-2.2)>0.001){

	// regular gamma correction
	r = pow(r, 1.0/gv->gamma);
	g = pow(g, 1.0/gv->gamma);
	b = pow(b, 1.0/gv->gamma);
      }
      else if (fabs(gv->gamma-1.0)>0.001){
	// sRGB standard gamma correction (similar to regular w/2.2)
	r = r<=0.0031308 ? 12.92*r : 1.055*pow(r, 1.0/2.4) - 0.055;
	g = g<=0.0031308 ? 12.92*g : 1.055*pow(g, 1.0/2.4) - 0.055;
	b = b<=0.0031308 ? 12.92*b : 1.055*pow(b, 1.0/2.4) - 0.055;
      }
	
      currMaterial->diffuseColor.set1Value(j, SbColor(r,g,b));
      
    } // end verts loop
    // end tone mapping
  } // end paths loop

  // end delay measure
  gettimeofday(&end, NULL);
  int ms = delay(start, end);
  if (!gv->silent)
    print("\a");
  print("Done. (%d ms, %f s, %f m, %f h)\n", ms, ms/1000.0, ms/1000.0/60.0,
	ms/1000.0/60.0/60.0);
  gv->times.render = ms;
}



//'''''''''''''''''WW'''''''''''WW''W''''''''''''WW'''''''
//                 WW           WW  W            WW       
//W WW  WW  WWW   WWW  WW  W WW WW  W  WW  W WW WWWW  WWWW
//WW   WWWW WW W W WW WWWW WW   WW  W WWWW WW    WW  WWW  
//WW   WW   WW W W WW WW   WW   WW W  WW   WW    WW    WWW
//WW    WWW WW W  WWW  WWW WW    WW    WWW WW    WW  WWWW 
//                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''

void Pane::renderVerts(thread_context *ctx){

  typedef struct renderVert_st {
    int numVerts;
    SoMaterial *mat;
    SoVertexProperty *vprop;
    SbMatrix *matrix;
    SbMatrix *matInvTran;
    float *okayNormals;
    int flags;
    SoIndexedShape *shape;
    int numCoordIdxs;
    int reverseOrdering;
  } renderVert_st_t;

  renderVert_st_t *passingVals = (renderVert_st_t *) ctx->other;
  int numVerts = passingVals->numVerts;
  SoMaterial *mat = passingVals->mat;
  SoVertexProperty *vprop = passingVals->vprop;
  SbMatrix *matrix = passingVals->matrix;
  SbMatrix *matInvTran = passingVals->matInvTran;
  float *okayNormals = passingVals->okayNormals;
  int flags = passingVals->flags;
  SoIndexedShape *shape = passingVals->shape;
  int numCoordIdxs = passingVals->numCoordIdxs;
  int reverseOrdering = passingVals->reverseOrdering;

  // loop over triangles
  int v[3];
  v[0] = v[1] = v[2] = -1;
  int triCount = 0;
  int numTris = 0;
  int colinearVerts = 0;
  for (int k=0; k<numCoordIdxs; k++){
    v[0] = v[1];
    v[1] = v[2];
    v[2] = shape->coordIndex[k];
    if (v[0] >= 0 && v[1] >= 0 && v[2] >= 0){
      	
      // triangle found
      numTris++;       // for actual counting
      triCount++;      // for even/odd triangle flipping
      if (ctx->id==0 && numTris%1000==0)
	print("\r%6.2f%%    ", k/(float)(numCoordIdxs)*100.0);

      // detect same vertices
      if (v[0] == v[1] || v[0] == v[2] || v[1] == v[2]){
	if (colinearVerts++ == 0)
	  cerr << "Warning! Colinear vertices in triangle!\n";
	continue;
      }

      // get points
      SbVec3f vert[3];
      matrix->multVecMatrix(vprop->vertex[v[0]], vert[0]);
      matrix->multVecMatrix(vprop->vertex[v[1]], vert[1]);
      matrix->multVecMatrix(vprop->vertex[v[2]], vert[2]);

      // get normal and area
      SbVec3f normal;
      SbVec3f edge1 = vert[1] - vert[0];
      SbVec3f edge2 = vert[2] - vert[0];
      
      if (reverseOrdering == 0)
	normal = edge1.cross(edge2);
      else
	normal = edge2.cross(edge1);
      
      // flip odd triangles if indexedtrianglestripset
      if (shape->isOfType(SoIndexedTriangleStripSet::getClassTypeId()) &&
	  triCount%2==0)
	normal = -normal;
      
      float area2 = normal.normalize();  

      // skip small trianglesn
      if (area2<0.00001){
        // cerr << "area too small, skipping.\n";
	continue;
      }

      // sample triangles vertices
      for (int i=0; i<3; i++){

	// skip points with no valid normal or already computed normal
	int j = v[i];
 	if (j%gv->numThreads != ctx->id)
 	  continue;
	
	if (okayNormals[j] >= 0.0  &&          // normal is defined
	    ! (area2 > 5.0*okayNormals[j] ))   // tri is not 5x bigger than last tri
	  continue;
	okayNormals[j] = area2;

	// set vert to 99% of one corner of triangle
	int i2 = (i+1)%3;
	int i3 = (i+2)%3;
	float w[3] = // { 1.0/3.0, 1.0/3.0, 1.0/3.0};
	  { 0.96, 0.02, 0.02 };
	SbVec3f cvert = vert[i]*w[0] + vert[i2]*w[1] + vert[i3]*w[2];

// 	SbVec3f normal2;
// 	matInvTran->multDirMatrix(vprop->normal[j], normal2);
// 	normal2.normalize();
	SbVec3f normal2 = normal;

	SbVec3f antinormal = -normal2;
	for (int c=0; c<3; c++){
	  if (normal2[c] == 0.0)
	    normal2[c] = 0.0;
	  if (antinormal[c] == 0.0)
	    antinormal[c] = 0.0;
	}

	// debugging
	// a = .5*b*h   h = sqrt(a*2)
	float offsetLen = sqrt(area2*2.0)*0.01;
	SbVec3f startPt = cvert+normal2*offsetLen;
	Ray r(startPt, antinormal, ctx, Ray::NO_PRIM, 1);
// 	if (ctx->mydrand48()<0.01)
// 	  r.show(2.0);

	//r.show2(offsetLen, SbColor(1,1,0));
	//showPhoton(cvert, SbColor(0,1,1));
	//showPhoton(startPt, SbColor(1,0,0));
	
	//continue;

	unsigned long long seed = ctx->mydrand48.seed;	
	SbColor col = getRadiancePM(r, flags);

	if (STOP_RENDERING){
	  if (gv->numThreads > 1)
	    pthread_exit(0);
	  else
	    return;
	}

	/*
 	if (col[0]==0.0 && col[1]==0.0 && col[2]==0.0){
	  while (1){
	    cerr << "BLACK\n";
	    ctx->mydrand48.seed = seed;

	    showPhoton(vert[i], SbColor(0.5,0,0));
	    showPhoton(vert[i2], SbColor(0,0.5,0));
	    showPhoton(vert[i3], SbColor(0,0,0.5));
	    
	    DEBUG_STATE=1;
	    col = getRadiancePM(r, flags, NULL);

	    STOP_RENDERING=1;
	    if (gv->numThreads > 1)
	      pthread_exit(0);
	    else
	      return;
	    
	  }
	}
	*/
	if (r.t>offsetLen*1.1){
	  cerr << "  MISSED: " << r.t << "\n";
	  
	  continue;
// 	  showPoint(r.pos, 0.5, SbColor(1,0,0));
// 	  Ray(r.x,normal,r.ctx).show(1.0, SbColor(0.0, 1.0, 0.0));     // green
// 	  Ray(r.x,normal2,r.ctx).show(1.0, SbColor(0.0, 1.0, 1.0));    // cyan
	}

	mat->diffuseColor.set1Value(j, col);
      }
      
    }
    else
	triCount = 0;
  } // end coords/triangle loop
  
  
  // completion meter
  completion[ctx->id] = 100.0;
  print("\r");
  for (int i=0; i<gv->numThreads; i++)
    print("%6.2f%%%c ", completion[i], completion_stage[i]);

  // time statistics
  struct tms renderTimes;
  times(&renderTimes);
  ctx->ticks += renderTimes.tms_utime + renderTimes.tms_cutime;

  if (gv->numThreads > 1)
    pthread_exit(0);
  else
    return;
}



//'''''''''''''WW''WW'''''''WWWW'''''''WW'WW'
//                 WW        WW       WW   WW
//WW    W W WW WW WWWW  WW   WW  WW W WW   WW
//WW WW W WW   WW  WW  WWWW  WW  WW W WW   WW
//WW WW W WW   WW  WW  WW    WW  WW W WW   WW
// WW WW  WW   WW  WW   WWW WWWW  WW   WW WW 
//                                           
//'''''''''''''''''''''''''''''''''''''''''''
//
// Writes Inventor scene to disk, inserting
// a comment.
//
void Pane::writeIv(){

  // prepend command line to scene
  static int addedCmdLine = 0;
  if (addedCmdLine == 0){
    addedCmdLine = 1;

    string out = createComment();
    
    SoInfo *info = new SoInfo();
    info->string.setValue(SbString(out.c_str()));
    gv->scene->insertChild(info, 0);
  }

  if (gv->figureMode)
    writeIvFile(gv->ivOutFile, scnRoot);        // arrows, photons
  else
    writeIvFile(gv->ivOutFile, gv->scene);      // just scene

  // write file nodes out
  if (1){

    // insert number before extension
    string fn;
    string::size_type lastPeriod = gv->ivOutFile.find_last_of('.');
    if (lastPeriod == string::npos)
      fn = gv->ivOutFile + ".";
    else
      fn = gv->ivOutFile.substr(0,lastPeriod+1);
    
    for (list <fileNode_t>::iterator p = fileNodeList.begin();
	 p != fileNodeList.end(); p++){
      string fileName = fn + p->name;
      writeIvFile(fileName.c_str(), p->copy);
    }
  }
  
}


void Pane::writeIvFile(const string &fileName, SoGroup *scene){

  // back up old file if it exists
  FileUtils::checkForOverwrite(fileName, gv->createBackup);
  
  SoOutput *out = new SoOutput();
  out->setBinary(TRUE);
  if (!out->openFile(fileName.c_str())){
    cerr << "Error: Could not open " << fileName << " for writing!\n";
    exit(1);
  }

  cerr << "Writing " << fileName << " ..." << flush;

  SoWriteAction *wa = new SoWriteAction(out);
  wa->apply((SoNode*)scene);
  out->closeFile();

  cerr << "done." << endl;

}

// replaces all SoFile nodes with editable (searchable) children
// This is needed to search SoFile nodes and update their normals, etc.
void Pane::makeEditableChildren(SoNode *scene){

  SoSearchAction sa;
  sa.setType(SoFile::getClassTypeId(), TRUE);
  sa.setSearchingAll(FALSE);
  sa.setFind(SoSearchAction::TYPE);
  sa.setInterest(SoSearchAction::ALL);
  sa.apply(scene);
  SoPathList plist = sa.getPaths();

  int len = plist.getLength();

  // record file nodes
  for (int i=0; i<len; i++){

    SoPath *path = (SoPath*) plist[i];
    SoFile *file = (SoFile *)path->getTail();

    // store information for SoFile for writing out node later
    fileNode_t fileNode;
    fileNode.parent = (SoGroup *)path->getNodeFromTail(1);

    // save file node name for output
    fileNode.name = file->name.getValue().getString();
    
    // create replacement for file node
    fileNode.copy = file->copyChildren();
    
    // replace parent's SoFile child with copy
    // replaceChild() refs replace, and unrefs (and possibly deletes) file    
    fileNode.parent->replaceChild(file, fileNode.copy);  
    
    fileNodeList.push_back(fileNode);
  }
  
}
