#ifndef PREVIEWQUAD_H
#define PREVIEWQUAD_H

#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoTexture2.h>

class SoSeparator;
class SoTexture2;
class SoCamera;
class SbColor;
class GlobalVars;
class SoXtExaminerViewer;

class PreviewQuad {
private:
  SoSeparator *previewSep;       // contains preview of rendered image
  SoTexture2 *previewTexture;    // preview texture pointer
  SbVec2s previewSize;           // actual size of the the texture buffer
  int previewRedraw;             // flag to signal refresh is wanted
  SoCamera *previewCam;          // preview camera node

  SbColor *previewBufFloat;      // more accurate intermediate preview buffer
  unsigned char *previewBufByte; // preview render buffer
  unsigned char *previewIter;    // buffer to store last refinement iteration
  float *previewSum;             // running sum of pixel brightness, per thread
  int *previewCnt;               // number of pixels in running sum, per thread
  int *previewDrawnCnt;          // number of pixels drawn (can be >pix), per thread
  
  float previewMult;             // scale to use when computing
                                 // float->byte
  int imgDim[2];
  int numThreads;
  bool constructed:1;
  bool enabled:1;
  
public:

  PreviewQuad();
  ~PreviewQuad();
  
  // call constructScene first, then preRender before every render.
  SoSeparator* constructScene(int *imgDim, int numThreads);
  void preRender(bool resetTexture=true);

  SoCamera* getCamera(){ return previewCam; }
  void setMult(float m){ previewMult=m; }
  float getMult() const { return previewMult; }
  bool getRedraw() const { return previewRedraw; }
  void setRedraw(bool b){ previewRedraw=b; }
  void finishTexture(){ previewTexture->image.finishEditing(); }
  void renderIfReady(SoXtExaminerViewer *xtViewer);

  // Start counters over for new image being drawn
  void resetCounters(int tId){
    previewSum[tId] = 0.0f;
    previewCnt[tId] = 0;
  }
  
  // copies (full) preview buffer into texture map, with cheap tonemapping
  void redrawPreview(const int *lowerLeft,
		     const int *upperRight);

  // updates tone mapping scalar, calls redrawPreview
  // called after each pass and when render completes.
  void updatePreview(const int *lowerLeft,
		     const int *upperRight);

  // fast, full copy from framebuffer to preview buffer
  void copyPixToPreview(int *lowerLeft, int *upperRight,
                        SbColor *pix);

  // copy framebuffer to preview buffer for PPM
  void copyPixToPreviewPPM(int *lowerLeft, int *upperRight,
                        SbColor *pix, int subpix);
  
  // draws big block into preview buffer
  void setPreviewPix(int x, int y, int tId, const SbColor &col,
                     int size, int iter,
                     int *lowerLeft, int *upperRight);
};


#endif
