#include "Main.h"
#include "Primitive.h"
#include "debug.h"

const SbColor Primitive::Black(0,0,0);
const SbColor Primitive::White(1,1,1);
const SbVec3f Primitive::XDir(1,0,0);
const SbVec3f Primitive::YDir(0,1,0);
const SbVec3f Primitive::ZDir(0,0,1);
const PrimDerivs Primitive::DefaultDerivs(XDir, YDir,
                                          SbVec3f(0,0,0), SbVec3f(0,0,0));
Material Primitive::defaultMat;

// bool Primitive::intersectCube(const SbVec3f &min, const SbVec3f &max) const {

//   SbVec3f myMin = getMin();
//   SbVec3f myMax = getMax();

//   return
//     myMin[0] < max[0] && myMax[0] > min[0] &&
//     myMin[1] < max[1] && myMax[1] > min[1] &&
//     myMin[2] < max[2] && myMax[2] > min[2];
// }

bool Primitive::intersectCube(const double min[3], const double max[3]) const {

  SbVec3f myMin = getMin();
  SbVec3f myMax = getMax();

  return
    myMin[0] <= max[0] && myMax[0] >= min[0] &&
    myMin[1] <= max[1] && myMax[1] >= min[1] &&
    myMin[2] <= max[2] && myMax[2] >= min[2];
}

bool Instance::intersectRay(Ray &r, double t0, double t1){
  SbVec3f post, dirt;
  mi.multVecMatrix(r.pos, post);
  mi.multDirMatrix(r.dir, dirt);
  float scale = dirt.normalize(); // newlen/oldlen
  
  Ray rt(post, dirt, r, r.ignorePrim);

  if (prim->intersectRay(rt, scale*t0, scale*t1)){

    SbVec3f p;
    m.multVecMatrix(rt.point(), p);

    // hit
    r.t = rt.t/scale;
    r.prim = this;
    r.prim2 = rt.prim2;
    r.prim2.push_back(rt.prim); // need to save hit prim
    r.bc = rt.bc;
    return true;
  }

  // no hit
  r.t  = -1;
  r.prim = NULL;
  r.bc = SbVec3f(1,0,0);
  return false;
}

namespace {
inline void eightCorners(const SbVec3f &min, const SbVec3f &max,
                  SbVec3f *x){
  x[0] = SbVec3f(min[0], min[1], min[2]);
  x[1] = SbVec3f(min[0], min[1], max[2]);
  x[2] = SbVec3f(min[0], max[1], min[2]);
  x[3] = SbVec3f(min[0], max[1], max[2]);
  x[4] = SbVec3f(max[0], min[1], min[2]);
  x[5] = SbVec3f(max[0], min[1], max[2]);
  x[6] = SbVec3f(max[0], max[1], min[2]);
  x[7] = SbVec3f(max[0], max[1], max[2]);
}

inline SbVec3f minBBox(const SbVec3f &min, const SbVec3f &max,
                const SbMatrix &m) {
  SbVec3f gmin(FLT_MAX, FLT_MAX, FLT_MAX);
  if (min[0]>max[0])
    return gmin;
  SbVec3f x[8];
  eightCorners(min, max, x);
  for (int i=0; i<8; i++){
    m.multVecMatrix(x[i], x[i]);
    gmin = BBox::vecMin(gmin, x[i]);
  }

  return gmin;
}

inline SbVec3f maxBBox(const SbVec3f &min, const SbVec3f &max,
                const SbMatrix &m) {
  SbVec3f gmax(-FLT_MAX, -FLT_MAX, -FLT_MAX);
  if (min[0]>max[0])
    return gmax;
  SbVec3f x[8];
  eightCorners(min, max, x);
  for (int i=0; i<8; i++){
    m.multVecMatrix(x[i], x[i]);
    gmax = BBox::vecMax(gmax, x[i]);
  }

  return gmax;
}
} // end anon namespace

SbVec3f  Instance::getMin() const {
  return minBBox(prim->getMin(), prim->getMax(), m);
}
SbVec3f  Instance::getMinShiny() const {
  return minBBox(prim->getMinShiny(), prim->getMaxShiny(), m);
}
SbVec3f  Instance::getMax() const {
  return maxBBox(prim->getMin(), prim->getMax(), m);
}
SbVec3f  Instance::getMaxShiny() const {
  return maxBBox(prim->getMinShiny(), prim->getMaxShiny(), m);
}
