
#ifndef SKYCLASS
#define SKYCLASS

#include "Main.h"
#include "GlobalVars.h"
#include "Ray.h"
#include "Pane.h"
#include "EnvironMap.h"
#include "Spectrum.h"

class Sky : public EnvironMap {

  SbVec3f sunBasisU, sunBasisV, sunBasisW;  // W points toward sun
  float sunRadius, sunRadiusAngle, sunDist, sunOmega; // (omega is solid angle)
  float turbidity, thetaSolar, thetaSolarSafe;
  SbColor sunRadiance;
  
  private:

  float AY,BY,CY,DY,EY,Ax,Bx,Cx,Dx,Ex,Ay,By,Cy,Dy,Ey;
  float Yz,xz,yz;
  EnvironMap sky;  // don't rely exclusively on subclass because virtual method calls are
                   // tricky, and we want different rotation matrices anyway
         
  SbColor B_m_0, B_p_0; // extinction
  SbColor *S_0_m;       // in-scattering (gas/Rayleigh)
  SbColor *S_0_p;       // in-scattering (haze/Mie)
  static Spectrum neta[1801];
  static int ti_S0pre, pi_S0pre;  // number of samples in theta/phi for S_0 cache
  static int ti_Lpre, pi_Lpre;    // number of hemisphere samples for int. L
  
  public:
  
  Sky();
  virtual ~Sky();
  
  virtual void init(GlobalVars *gv) { init(gv, 17.0); }
  virtual void init(GlobalVars *gv, float time);
  virtual SbColor impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple=0);
  //virtual SbColor volImpSample(Ray &r, int simple);  
  virtual float getSample(double eps1, double eps2,
                          SbVec3f &dir, SbColor &c);
  virtual SbColor volImpSample(Ray &r, int samples);
  virtual SbColor getRadiance(SbVec3f &dir) { return getRadiance(dir, true); }
  virtual SbColor getRadiance(SbVec3f &dir, bool includeSun);
  virtual SbColor getEmissiveFlux(float receiverArea);

  void ComputeOutScatteringCoeffs();
  void ComputeInScatteringCoeffs();
  void ComputeInScatteringForOneDir(const SbVec3f &out,
                                    SbVec3f &S_0_m_theta, SbVec3f &S_0_p_theta);
  
  SbColor ComputeAtten(double h0, double cosT, double t);

  SbColor SolveInScatter(double cosT,
                        double a_haze, const SbColor &K1, double H1, double u1s, const SbColor &t_haze, const SbVec3f &B_p,
                        double a_mol,  const SbColor &K2, double H2, double u2s, const SbColor &t_mol, const SbVec3f &B_m);
  
  void ComputeAttenAndInScatter(double h0, double cosT, double phi, double t,
                                SbColor &atten, SbColor &inscatter);
  
  SbColor ComputeAerialPersp(const Ray &r, SbColor &c);
  
};

#endif /* ENVIRONMAPCLASS */
  
