//WWWWWW'''''''WW''''''''''''''''WWW'''''''''''''''WWW''''''''''''''''''
//  WW                            WW                WW                  
//  WW    W WW WW  WWW WWW   WWW  WW   WW      WWW  WW   WWW  WWWW  WWWW
//  WW    WW   WW W WW WW W W WW  WW  WWWW    WW W  WW  W WW WWW   WWW  
//  WW    WW   WW W WW WW W W WW  WW  WW      WW    WW  W WW   WWW   WWW
//  WW    WW   WW  WWW WW W  WWW WWWW  WWW     WWW WWWW  WWW WWWW  WWWW 
//                            WW                                        
//''''''''''''''''''''''''''WWW'''''''''''''''''''''''''''''''''''''''''

#ifndef TRIANGLECLASS
#define TRIANGLECLASS

#include <float.h>
#include <vector>

#include "Main.h"
#include "Misc.h"
#include "Primitive.h"
#include "Material.h"
#include "SoMatPlus.h"
#include "Ray.h"
#include "intersect_ray_tri.h"
#include "intersect_tri_cube.h"
#include "Misc.h"
#include "Texture.h"

using namespace std;

class Texture;
class Pane;
class SoPrimitiveVertex;

class TriMesh {
  
public:

  static vector<TriMesh*> meshs;
  static void initMeshs() { }
  static void deleteMeshs() {
    for (int i=0; i<meshs.size(); i++){
      delete meshs[i]->texture;
      delete meshs[i];
    }
  }

  // This computes internal normals for each triangle mesh.
  // See alternative computeNormalsIv() for updating the
  // inventor scene graph directly.
  static void computeAllNormals(){
    cerr << "Computing normals for all Triangle meshes.\n";
    for (int i=0; i<meshs.size(); i++)
      meshs[i]->computeNormals();
  }
  
  vector<int> indexes;       // three indexes == triangle
  vector<int> normalIndexes;
  vector<SbVec3f> verts;
  vector<SbColor> diffuse;
  vector<SbVec3f> normals;
  vector<SbVec2f> uvs;
  Texture *texture;
  Material *mat;

  TriMesh() : texture(NULL), mat(NULL) { }

  int add(const SbVec3f &v0, const SbVec3f &v1, const SbVec3f &v2,
          const SbColor &d0, const SbColor &d1, const SbColor &d2,
          const SbVec3f &n0, const SbVec3f &n1, const SbVec3f &n2){

    int idx = verts.size();
    verts.push_back(v0);
    verts.push_back(v1);
    verts.push_back(v2);
    diffuse.push_back(v0);
    diffuse.push_back(v1);
    diffuse.push_back(v2);
    normals.push_back(v0);
    normals.push_back(v1);
    normals.push_back(v2);
    indexes.push_back(idx+0);
    indexes.push_back(idx+1);
    indexes.push_back(idx+2);
    return idx;
  }
  int add(){
    SbColor z(0,0,0);
    SbColor u(0,1,0);
    return add(z,z,z, z,z,z, u,u,u);
  }
  int numTris() const { return indexes.size()/3; }
  int numVerts() const { return verts.size(); }

  void computeNormals();

};

class Triangle : public Primitive {
  
protected:

  static int errorTris;
  static float emissiveArea;
  
  TriMesh *mesh;
  int n; // index to beginning of 3 vertex indices
  
private:
  
public:

  Triangle();
  Triangle(TriMesh *mesh_, int n_) :
    mesh(mesh_), n(n_) { }
  Triangle(Pane *pane, SoCallbackAction *action,
	   const SoPrimitiveVertex *v0,
	   const SoPrimitiveVertex *v1,
	   const SoPrimitiveVertex *v2,
           bool &okay) {
    createTriangle(pane, action, v0, v1, v2, okay);
  }
  void init();
  void createTriangle(Pane *pane, SoCallbackAction *action,
                      const SoPrimitiveVertex *v0_,
                      const SoPrimitiveVertex *v1_,
                      const SoPrimitiveVertex *v2_,
                      bool &okay);
  
  virtual ~Triangle();
  
  // returns vertex for triangle n, vertex vert
  // FIXME - replace n with vertIndex* already pointed to it
  bool hasDiffuse() const { return !mesh->diffuse.empty(); }
  bool hasNormals() const { return !mesh->normals.empty(); }
  bool hasIndexedNormals() const { return !mesh->normalIndexes.empty(); }
  const SbVec3f& vert(int v) const { return mesh->verts[mesh->indexes[n+v]];  }
  const SbColor& diffuse(int v) const { return mesh->diffuse[mesh->indexes[n+v]]; }
  const SbVec3f& normals(int v) const {
    return hasIndexedNormals() ?
      mesh->normals[mesh->normalIndexes[n+v]] :
      mesh->normals[mesh->indexes[n+v]];
  }
  SbVec3f& vert(int v){ return mesh->verts[mesh->indexes[n+v]]; }
  SbColor& diffuse(int v){ return mesh->diffuse[mesh->indexes[n+v]]; }
  SbVec3f& normals(int v){
    return hasIndexedNormals() ?
      mesh->normals[mesh->normalIndexes[n+v]] :
      mesh->normals[mesh->indexes[n+v]];
  }
  const SbVec2f& uv(int v) const { return mesh->uvs[mesh->indexes[n+v]]; }
  const Texture* texture() const { return mesh->texture; }

  // Primitive virtual functions
  PrimType getType() const { return PRIMTYPE_TRIANGLE; }
  virtual string getTypeName() const { return "Triangle"; }
  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8){

    r.ctx->isects++;

    // initialize as no intersection
    r.t = -1;
    r.prim = NULL;
    
    double t, u=0, v=0;
    if (this == r.ignorePrim)
      return false;
    
    if (!intersect_ray_tri(r.pos, r.dir, vert(0), vert(1), vert(2), &t, &u, &v))
      return false;

    if (t<t0)
      return false;

    if (t>t1)
      return false;

    // store triangle distance, pointer, barycentric coords
    r.t = t;
    r.prim = this;
    r.bc[0] = 1.0f-u-v; // FIXME: can be delayed
    r.bc[1] = u;
    r.bc[2] = v;
    
    return true;
  }

  bool intersectCube(const SbVec3f &min, const SbVec3f &max) const {
      double boxcenter[3], boxhalfsize[3];
      boxcenter[0] = (max[0]+min[0])*0.5;
      boxcenter[1] = (max[1]+min[1])*0.5;
      boxcenter[2] = (max[2]+min[2])*0.5;
      boxhalfsize[0] = max[0]-boxcenter[0];
      boxhalfsize[1] = max[1]-boxcenter[1];
      boxhalfsize[2] = max[2]-boxcenter[2];
      return intersect_tri_cube(boxcenter, boxhalfsize,
                                vert(0), vert(1), vert(2));
  }
  bool intersectCube(const double min[3], const double max[3]) const {
      double boxcenter[3], boxhalfsize[3];
      boxcenter[0] = (max[0]+min[0])*0.5;
      boxcenter[1] = (max[1]+min[1])*0.5;
      boxcenter[2] = (max[2]+min[2])*0.5;
      boxhalfsize[0] = max[0]-boxcenter[0];
      boxhalfsize[1] = max[1]-boxcenter[1];
      boxhalfsize[2] = max[2]-boxcenter[2];
      return intersect_tri_cube(boxcenter, boxhalfsize,
                                vert(0), vert(1), vert(2));
  }
  bool intersectCubeCenterHalf(const double center[3],
                               const double half[3]) const {
    return intersect_tri_cube(center, half, vert(0), vert(1), vert(2));
  }

  SbVec3f getMin() const {
    SbVec3f min(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int v=0; v<3; v++)
      min = BBox::vecMin(min, vert(v));
    return min;
  }
  SbVec3f getMinShiny() const {
    return getMat()->isShiny() ?
      getMin() : SbVec3f(FLT_MAX, FLT_MAX, FLT_MAX);
  }
  
  SbVec3f getMax() const {
    SbVec3f max(-FLT_MAX, -FLT_MAX, -FLT_MAX);
    for (int v=0; v<3; v++)
      max = BBox::vecMax(max, vert(v));
    return max;
  }
  SbVec3f getMaxShiny() const {
    return getMat()->isShiny() ?
      getMax() : SbVec3f(-FLT_MAX, -FLT_MAX, -FLT_MAX);
  }
  float getMin(int dim) const {
    return min3(vert(0)[dim], vert(1)[dim], vert(2)[dim]);
  }
  float getMax(int dim) const {
    return max3(vert(0)[dim], vert(1)[dim], vert(2)[dim]);
  }
  
  Material* getMat() { return mesh->mat; }
  const Material* getMat() const { return mesh->mat; }
  SbColor getIntDiffColor(const Ray &r) const;
  SbColor getIntEmission(const Ray &r) const;
  virtual SbColor  getEmission()  const { return getMat()->emission; }
  SbVec3f getIntNormal(   const SbVec3f &bc) const { return getIntNormal(bc, NULL);}
  SbVec3f getIntNormal(   const SbVec3f &bc, const PrimDerivs *pDer) const;
  SbVec3f getNormal()                        const;  
  SbVec3f getNormal(      const SbVec3f &bc) const { return getNormal(); }
  bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer) const;
  
  void calcBaryCoords(const SbVec3f &pnt, SbVec3f &bc) const;
  void useBaryCoords(const double a, const double b, const double c,
		     SbVec3f &bc);
  
  // draw verts (spheres) + tri (polygon)
  void show(SbColor c = SbColor(1.0, 1.0, 1.0));

  // draw as edge lines, slightly inset
  void show2();

  // temporary
  void getSTfromBC(const SbVec3f &bc, float &s, float &t) const {
    
    // calculate image coordinates (interpolate)
    SbVec2f st = bc[0]*uv(0) + bc[1]*uv(1) + bc[2]*uv(2);
    s = st[0];
    t = st[1];
  }
  SbColor getColor(const SbVec3f &bc) const {
    float s, t;
    getSTfromBC(bc, s, t);
    return texture()->getColor(s, t);
  }
  SbColor getColor(const SbVec3f &bc, const SbColor &diffuse) const {
    float s, t;
    getSTfromBC(bc, s, t);
    return texture()->getColor(s, t, diffuse);
  }
  SbVec3f getDeriv(const SbVec3f &bc, float du=0, float dv=0) const {
    float s, t;
    getSTfromBC(bc, s, t);
    return texture()->getDeriv(s, t, du, dv);
  }

  // display statistical information
  static void printGlobalStats();

};

ostream& operator<< (ostream& os, const Triangle& x);

class EmissiveTriangle : public Triangle {

  private:

  // projection map stuff
  static const int mapSize[2];
  char *map;                 // value of NULL means don't use map
  float *celluv;             // precomputed u,v values of filled (only) cells
  float mapDeltaCDF;         // delta per filled box in map CDF
  float mapPDF;              // distribution of filled boxes
  int cells;

  public:
  
  // precomputed to accelerate direct lighting calculation
  // or only needed for lights
  float area;
  float emisAvg;
  SbVec3f normal;
  SbColor emission;          // Could be stored with shader (only needed for lights)
                             // Emissive flux.
  ONB onb;                   // ortho-normal basis of center of triangle

  EmissiveTriangle();
  EmissiveTriangle(Pane *pane, SoCallbackAction *action,
		   const SoPrimitiveVertex *v0,
		   const SoPrimitiveVertex *v1,
		   const SoPrimitiveVertex *v2,
                   bool &okay) :
    Triangle(pane, action, v0, v1, v2, okay), map(NULL), celluv(NULL) {
    if (!okay) return;
    init(pane);
  }
  EmissiveTriangle(Pane *pane, TriMesh *mesh_, int n_) :
    Triangle(mesh_, n_) {
    init(pane);
  }
  
  virtual ~EmissiveTriangle(){
    delete [] map;
    delete [] celluv;
  }
  void init(Pane *pane);

  
  PrimType getType() const { return PRIMTYPE_TRIANGLE | PRIMTYPE_EMISTRIANGLE; }
  virtual string getTypeName() const { return "EmissiveTriangle"; }
  
  float    getEmisAvg()   const { return emisAvg;  }
  virtual SbColor  getEmission()  const { return emission; }
  SbVec3f  getNormal()    const { return normal;   }
  SbVec3f  getNormal(const SbVec3f &bc) const { return getNormal(); }
  const SbVec3f& getNormalPre() const { return normal;   } // precomputed normal

  int UseMap(){ return map?true:false; }
  void genImpMap(const SbVec3f &center, float rad);
  void sampleDir(SbVec3f &dir, float &pdf, float e1, float e2);
  void printMap();
  
};

ostream& operator<< (ostream& os, const EmissiveTriangle& x);

class DMTri : public Triangle {

public:

  float maxHeight;
  
  DMTri();
  
  DMTri(Pane *pane, SoCallbackAction *action,
        const SoPrimitiveVertex *v0,
        const SoPrimitiveVertex *v1,
        const SoPrimitiveVertex *v2,
        bool &okay);
  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);
  bool intersectRay2(Ray &r, double t0=1e-5, double t1=1e8);

  SbVec3f getMin() const {
    SbVec3f min(FLT_MAX, FLT_MAX, FLT_MAX);
    for (int v=0; v<3; v++){
      min = BBox::vecMin(min, vert(v));
      min = BBox::vecMin(min, vert(v)+maxHeight*normals(v));
    }
    return min;
  }

  SbVec3f getMax() const {
    SbVec3f max(-FLT_MAX, -FLT_MAX, -FLT_MAX);
    for (int v=0; v<3; v++){
      max = BBox::vecMax(max, vert(v));
      max = BBox::vecMax(max, vert(v)+maxHeight*normals(v));
    }
    return max;
  }
  
  SbVec3f  getNormal(const SbVec3f &bc) const;
  SbColor  getIntDiffColor(const Ray &r) const;
};

#endif /* TRIANGLECLASS */
