//WWWW'''''''WW''''''''''''''''''WW''''''''''
//WW  W      WW                              
//WW  W  WW  WWW  W WW  WWW  WWW WW WWW   WWW
//WW  W WWWW WW W W WW W WW W WW WW WW W W WW
//WW  W WW   WW W W WW W WW W WW WW WW W W WW
//WWWW   WWW WWW   WWW  WWW  WWW WW WW W  WWW
//                       WW   WW           WW
//'''''''''''''''''''''WWW''WWW''''''''''WWW'

#include <Inventor/nodes/SoPointSet.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include <pthread.h>
#ifndef __darwin__
#include <execinfo.h>
#endif
#include <fenv.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#include "debug.h"
#include "Main.h"
#include "Misc.h"

using namespace std;

pthread_mutex_t lock_printVec = PTHREAD_MUTEX_INITIALIZER;

#ifndef __darwin__
extern bool getIsAuthor();
int ENABLED_FPES =
  getIsAuthor() && !getenv("PANE_DISABLE_FPE") ||
  getenv("PANE_ENABLE_FPE") ?
  FE_ALL_EXCEPT ^ FE_INEXACT ^ FE_UNDERFLOW : 0;
#endif

namespace {
  pthread_mutex_t lock_showPoint  = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_showCube   = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_showPhoton = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_print      = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_t lock_scnroot    = PTHREAD_MUTEX_INITIALIZER;
}

// debugging
void complain(int line){
  cerr << "failure at line " << line << "\n";
}

ostream& operator<< (ostream& os, const SbVec3f& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x[0] << ", " << x[1] << ", " << x[2] << ")";
  resetios(os);
  return(os);
}

ostream& operator<< (ostream& os, const SbMatrix& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "[" << x[0][0] << ", " << x[0][1] << ", " << x[0][2] << ", " << x[0][3] << "\n";
  os << " " << x[1][0] << ", " << x[1][1] << ", " << x[1][2] << ", " << x[1][3] << "\n";
  os << " " << x[2][0] << ", " << x[2][1] << ", " << x[2][2] << ", " << x[2][3] << "\n";
  os << " " << x[3][0] << ", " << x[3][1] << ", " << x[3][2] << ", " << x[3][3] << "]";
  resetios(os);
  return(os);
}

void showPoint(SbVec3f p, double radius, SbColor c, int clear){

  pthread_mutex_lock(&lock_showPoint);
  
  // add self to scene graph
  static SoSeparator *showPointSep = NULL;
  static SoLOD *lodSphere = NULL;
  if (showPointSep == NULL){
    
    // add point sep to scene
    showPointSep = new SoSeparator();
    scnRoot->addChild(showPointSep);

    // add simplifying complexity 
    //SoComplexity *comp = new SoComplexity();
    //comp->value.setValue(0.1);
    //showPointSep->addChild(comp);

    // create lod sphere
    lodSphere = new SoLOD();
    lodSphere->range.set1Value(0, 10);
    lodSphere->range.set1Value(1, 100);
    lodSphere->range.set1Value(2, 200);
    lodSphere->range.set1Value(3, 400);
    
    SoSeparator *sep = new SoSeparator();
    SoComplexity *comp = new SoComplexity();
    comp->value.setValue(.8);
    sep->addChild(comp);
    sep->addChild(new SoSphere());
    lodSphere->addChild(sep);

    sep = new SoSeparator();
    comp = new SoComplexity();
    comp->value.setValue(.5);
    sep->addChild(comp);
    sep->addChild(new SoSphere());
    lodSphere->addChild(sep);

    sep = new SoSeparator();
    comp = new SoComplexity();
    comp->value.setValue(0.25);
    sep->addChild(comp);
    sep->addChild(new SoSphere());
    lodSphere->addChild(sep);
    
    sep = new SoSeparator();
    comp = new SoComplexity();
    comp->value.setValue(0.1);
    sep->addChild(comp);
    sep->addChild(new SoSphere());
    lodSphere->addChild(sep);

    sep = new SoSeparator();
    lodSphere->addChild(sep);
  }

  if (clear){
    showPointSep->removeAllChildren();

    // add simplifying complexity 
    //SoComplexity *comp = new SoComplexity();
    //comp->value.setValue(0.1);
    //showPointSep->addChild(comp);
    
    pthread_mutex_unlock(&lock_showPoint);
    return;
  }

  // Ray debugging
  SoSeparator *sep = new SoSeparator();
  static double h=0.0;
  static int count = 0;
  double s=1.0, v=1.0;

  if (count++>1000000){
    pthread_mutex_unlock(&lock_showPoint);
    return;
  }
  h += 0.03;
  float myh = h;

  // Translate base
  SoTransform *trn = new SoTransform();
  //trn->scaleFactor.setValue(scale, scale, scale);
  trn->translation.setValue(p);
  sep->addChild(trn);

  // Give Color
  SoMaterial *mat = new SoMaterial();
  mat->shininess.setValue(0);
  mat->ambientColor.setValue(0,0,0);
  if (c[0]>=0.0){
    // for (int i=0; i<3; i++)
    // c[i] = 1.0-exp(-2.0*c[i]/10.0);
    mat->diffuseColor.setValue(c);
    //mat->emissiveColor.setValue(c);
  }
  else
    mat->diffuseColor.setHSVValue(myh, s, v);
  //mat->transparency.setValue(0.5);
  sep->addChild(mat);

  // Add sphere
  //SoSphere *sphr = new SoSphere();
  //sphr->radius.setValue(radius);
  //sep->addChild(sphr);

  // add scaled LOD sphere
  SoTransform *sphTran = new SoTransform();
  sphTran->scaleFactor.setValue(radius, radius, radius);
  sep->addChild(sphTran);
  sep->addChild(lodSphere);

  pthread_mutex_lock(&lock_scnroot);
  showPointSep->addChild(sep);
  pthread_mutex_unlock(&lock_scnroot);

  pthread_mutex_unlock(&lock_showPoint);
}

void showCube(SbVec3f min, SbVec3f max, SbVec3f c, int clear){

  // add self to scene graph
  static SoSeparator *showCubeSep = NULL;
  pthread_mutex_lock(&lock_showCube);
  if (showCubeSep == NULL){
    // add cube sep to scene
    showCubeSep = new SoSeparator();
    scnRoot->addChild(showCubeSep);

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showCubeSep->addChild(style);
  }
  else if (clear){
    showCubeSep->removeAllChildren();

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showCubeSep->addChild(style);
    
    pthread_mutex_unlock(&lock_showCube);
    return;
  }
  pthread_mutex_unlock(&lock_showCube);
  
  // Cube
  SoSeparator *cubeSep = new SoSeparator();
  SoMaterial *cubeMat = new SoMaterial();
  cubeMat->diffuseColor.setValue(c);
  cubeSep->addChild(cubeMat);
  
  /*     ____
       /|1  /|5    z
      / |  / |     |
     /__|0=min4    |__y
    2|  / |6=max  /
     | /  | /    x
    3|/___|7

  */
    
  float vertices[8][3] =
  {  { min[0], min[1], min[2]},   // 0
     { min[0], min[1], max[2]},   // 1
     { max[0], min[1], max[2]},   // 2
     { max[0], min[1], min[2]},   // 3
     { min[0], max[1], min[2]},   // 4
     { min[0], max[1], max[2]},   // 5
     { max[0], max[1], max[2]},   // 6
     { max[0], max[1], min[2]}    // 7
  };

  int32_t indices[40] =
  {
    0, 1, 2, 3,  -1,
    2, 6, 7, 3,  -1,
    4, 7, 6, 5,  -1,
    0, 4, 5, 1,  -1,
    1, 5, 6, 2,  -1,
    0, 3, 7, 4,  -1
  };

  // coordinates
  SoCoordinate3 *cubeCoords = new SoCoordinate3();
  cubeCoords->point.setValues(0,8, vertices);
  cubeSep->addChild(cubeCoords);

  // line set
  SoIndexedLineSet *cubeLineSet = new SoIndexedLineSet();
  cubeLineSet->coordIndex.setValues(0, 40, indices);
  cubeSep->addChild( cubeLineSet );

  // add to scene
  pthread_mutex_lock(&lock_showCube);
  showCubeSep->addChild(cubeSep);
  pthread_mutex_unlock(&lock_showCube);

  
}

void printCol(SbColor x){
  pthread_mutex_lock(&lock_printVec);
  cerr << "(";
  cerr << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  cerr << x[0] << ", ";
  cerr << x[1] << ", ";
  cerr << x[2] << ") " << flush;
  resetios(cerr);
  pthread_mutex_unlock(&lock_printVec);
  return;
}

void showPhoton(SbVec3f p, SbColor c, int clear){

  pthread_mutex_lock(&lock_showPhoton);
  static SoSeparator *showPhotSep = NULL;
  static int num = 0;
  static SoMaterial *mat = NULL;
  static SoVertexProperty *vProp = NULL;
  static SoPointSet *pntSet = NULL;

  // figure purposes
  //p[1] = 100.0;
  
  if (showPhotSep == NULL){

    // create sep
    showPhotSep = new SoSeparator();

    // make cubes unpickable
    SoPickStyle *style = new SoPickStyle();
    style->style.setValue(SoPickStyle::UNPICKABLE);
    showPhotSep->addChild(style);

    // add material 
    mat = new SoMaterial();
    showPhotSep->addChild(mat);

    // add draw style
    SoDrawStyle *drawStyle = new SoDrawStyle();
    drawStyle->pointSize.setValue(2.0); //10
    showPhotSep->addChild(drawStyle);
    
    // add PointSet with vertexProperty
    vProp = new SoVertexProperty();
    vProp->materialBinding = SoVertexProperty::PER_VERTEX;
    pntSet = new SoPointSet();
    pntSet->vertexProperty.setValue(vProp);
    pntSet->numPoints.setValue(num);
    showPhotSep->addChild(pntSet);
    
    // add phot sep to scene
    pthread_mutex_lock(&lock_scnroot);
    scnRoot->addChild(showPhotSep);
    pthread_mutex_unlock(&lock_scnroot);
  }

  if (clear){
    num = 0;
    vProp->vertex.setNum(num);
    pntSet->numPoints.setValue(num);

    pthread_mutex_unlock(&lock_showPhoton);
    return;
  }

  // add point
  if (num<1000000){
    mat->diffuseColor.set1Value(num,
				clamp(c[0], (float)0.0, (float)1.0),
				clamp(c[1], (float)0.0, (float)1.0),
				clamp(c[2], (float)0.0, (float)1.0));
    vProp->vertex.set1Value(num++, p);
    pntSet->numPoints.setValue(num);
  }
  
  pthread_mutex_unlock(&lock_showPhoton);
}

void print(char const *fmt, ...){
  va_list ap;
  va_start(ap, fmt);

  pthread_mutex_lock(&lock_print);
  vfprintf(stderr, fmt, ap);
  pthread_mutex_unlock(&lock_print);
  
  va_end(ap);
}


void drawHist(float *flux_map, int n_phi, int n_theta, float scale,
	      SbVec3f x, SbVec3f n){

  float phi, theta;
  SbVec3f neww;
  ONB onb(n);
  
  // display histogram
  static SoSeparator *histSep = NULL;
  int c_idx=0, f_idx=0;
  SoCoordinate3 *coord;
  SoIndexedFaceSet *faceSet;
  if (!histSep){
    histSep = new SoSeparator();
    histSep->addChild(new SoMaterial());
    coord = new SoCoordinate3();
    faceSet = new SoIndexedFaceSet();
    histSep->addChild(coord);
    histSep->addChild(faceSet);
    scnRoot->addChild(histSep);
  }
  else {
    coord = (SoCoordinate3 *) histSep->getChild(0);
    faceSet = (SoIndexedFaceSet *) histSep->getChild(1);
    faceSet->coordIndex.setNum(0);
    coord->point.setNum(0);
  }

  // add point to coord
  int centIdx = c_idx++;
  coord->point.set1Value(centIdx, x);
    
  for (int u=0; u<n_theta; u++)
    for (int v=0; v<n_phi; v++){

      //   ^     2 3
      // u |     
      // v 0->1  0 1
	
      // 0
      theta = acos(sqrt( u/(double)n_theta ));
      phi = v/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    sin(theta)*sin(phi),
		    cos(theta));
      neww = onb.toWorldSpace(neww);
      coord->point.set1Value(c_idx+0,x+flux_map[u*n_phi+v]*scale*neww);

      // 1
      theta = acos(sqrt( u/(double)n_theta ));
      phi = (v+1)/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    sin(theta)*sin(phi),
                    cos(theta));
      neww = onb.toWorldSpace(neww);
      coord->point.set1Value(c_idx+1,x+flux_map[u*n_phi+v]*scale*neww);

      // 2
      theta = acos(sqrt( (u+1)/(double)n_theta ));
      phi = v/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
                    sin(theta)*sin(phi),
                    cos(theta));
      neww = onb.toWorldSpace(neww);
      coord->point.set1Value(c_idx+2,x+flux_map[u*n_phi+v]*scale*neww);

      // 3
      theta = acos(sqrt( (u+1)/(double)n_theta ));
      phi = (v+1)/(double)n_phi*2.0*M_PI;
      neww.setValue(sin(theta)*cos(phi),
		    sin(theta)*sin(phi),
                    cos(theta));
      neww = onb.toWorldSpace(neww);
      coord->point.set1Value(c_idx+3,x+flux_map[u*n_phi+v]*scale*neww);


      // draw faces
      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, -1);

      faceSet->coordIndex.set1Value(f_idx++, centIdx);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, -1);
	       
      faceSet->coordIndex.set1Value(f_idx++, c_idx+1);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+3);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+2);
      faceSet->coordIndex.set1Value(f_idx++, c_idx+0);
      faceSet->coordIndex.set1Value(f_idx++, -1);
       
      c_idx += 4;
    }
}

namespace abi {
  extern "C" char *__cxa_demangle(const char *mangled_name,
                                  char *buf, size_t *n, int *status);
}

string& getProgramName() {
  static string name = "/dev/null";
  return name;
}

void backTrace(bool both){
#ifdef __darwin__
  return;
#else

  int num=1024;
  void *array[num];
  
  size_t size = backtrace(array, num);
  char **strings = backtrace_symbols(array, size);

  cerr << "\nBacktrace:\n\n";
  
  // mangled
  if (both){
    cerr << "(mangled)\n";
    for (int i=0; i<size; i++)
      cerr << strings[i] << "\n";

    cerr << "\n(demangled)\n";
  }

  // demangled
  for (int i=0; i<size; i++){

    // search for "_Z"
    string s = strings[i];
    string::size_type begin = s.find("_Z", 0);
    string::size_type end = s.find('+', begin);
    
    if (begin == string::npos || end == string::npos){
      cerr << strings[i] << "\n";
    } else {
      string sym = s.substr(begin, end-begin);

      int status = 0;
      char *name = abi::__cxa_demangle(sym.c_str(), NULL, NULL, &status);

      if (name){ cerr << s.replace(begin, end-begin, name) << "\n"; free(name); }
      else if (status == -1) cerr << "Out of memory\n";
      else if (status == -2) cerr << "Invalid mangled name: " << strings[i] << "\n";
      else if (status == -3) cerr << "Invalid arguments\n";
      else cerr << "Unexepected __cxa_demangle status: " << status << "\n";
    }
  }

  // line numbers
  string cmd = "addr2line -e " + getProgramName();
  for (int i=0; i<size; i++){

    // search for "["
    string s = strings[i];
    string::size_type begin = s.find("[", 0);
    string::size_type end = s.find(']', begin);
    
    if (begin == string::npos || end == string::npos){
      cmd = cmd + " 0x0";
    } else {
      cmd = cmd + " " + s.substr(begin+1, end-begin-1);
    }
  }
  cerr << "\n(executing: " << cmd << ")" << endl;
  int ret = system(cmd.c_str());
  if (ret==-1)
    cerr << "ERROR: executing " << cmd << "\n";
  else if (ret)
    cerr << "ERROR: command `" << cmd << "` returned " << ret << "\n";
  
  free(strings);
#endif  /* __darwin__ */
}

// catch signals and dump backtrace
void setupSignals(){
  struct sigaction mySigAction;
  bzero((void *) &mySigAction, sizeof(struct sigaction));
  mySigAction.sa_sigaction = sigactionHandler;
  mySigAction.sa_flags     = SA_SIGINFO;
  
  if (sigaction(SIGINT,  &mySigAction, NULL) ||
      sigaction(SIGSEGV,  &mySigAction, NULL) ||
      sigaction(SIGFPE,  &mySigAction, NULL) ||
      sigaction(SIGUSR1, &mySigAction, NULL) ||
      sigaction(SIGHUP,  &mySigAction, NULL) ) {
    perror("sigaction failed ");
    exit(-1);
  }
}

namespace sigactionHandlerNS {
  int imgHeight = 0;
  pthread_t threadIds[32];
  int pixel[32][2] = {
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
    { -1, -1 }, { -1, -1 }, { -1, -1 }, { -1, -1 },
  };
}

void sigactionHandler(int sig, siginfo_t *info, void *context) {

//   printf("Got signal %d, si_signo = %d, si_code = %d, si_addr = 0x%x\n",
// 	 sig, info->si_signo, info->si_code, info->si_addr);

  // turn off further signal handling
  for (int i=0; i<NSIG; i++)
    signal(i, SIG_IGN);
  
  cerr << "\n\nCAUGHT SIGNAL: ";
  string name;
  switch (sig) {
  case SIGHUP:  name="SIGHUP";  break;
  case SIGINT:  name="SIGINT";  break;
  case SIGQUIT: name="SIGQUIT"; break;
  case SIGILL:  name="SIGILL";  break;
  case SIGABRT: name="SIGABRT"; break;
  case SIGFPE:  name="SIGFPE";  break;
  case SIGKILL: name="SIGKILL"; break;
  case SIGSEGV: name="SIGSEGV"; break;
  case SIGPIPE: name="SIGPIPE"; break;
  case SIGALRM: name="SIGALRM"; break;
  case SIGTERM: name="SIGTERM"; break;
  case SIGUSR1: name="SIGUSR1"; break;
  case SIGUSR2: name="SIGUSR2"; break;
  case SIGCHLD: name="SIGCHLD"; break;
  case SIGCONT: name="SIGCONT"; break;
  case SIGSTOP: name="SIGSTOP"; break;
  case SIGTSTP: name="SIGTSTP"; break;
  case SIGTTIN: name="SIGTTIN"; break;
  case SIGTTOU: name="SIGTTOU"; break;
  default:      name="UNKNOWN"; break;
  }

  cerr << name << " !\n";

  if (sig == SIGFPE){
    string err;
    switch (info->si_code){
      case FPE_INTDIV: err="integer divide by zero           "; break;
      case FPE_INTOVF: err="integer overflow                 "; break;
      case FPE_FLTDIV: err="floating point divide by zero    "; break;
      case FPE_FLTOVF: err="floating point overflow          "; break;
      case FPE_FLTUND: err="floating point underflow         "; break;
      case FPE_FLTRES: err="floating point inexact result    "; break;
      case FPE_FLTINV: err="floating point invalid operation "; break;
      case FPE_FLTSUB: err="subscript out of range           "; break;
    }
      
    cerr << "Floating-point exception: " << err << endl;

    cerr << "\nTo disable floating-point exception detection, set \n"
      "environment variable PANE_DISABLE_FPE.\n";

    int threadId = -1;
    pthread_t selfId = pthread_self();
    using namespace sigactionHandlerNS;
    for (threadId=0; threadId<32; threadId++)
      if (threadIds[threadId] == selfId)
        break;
    if (threadId>=32){
      cerr << "Unknown thread ID.\n";
    } else {
      int i = pixel[threadId][0];
      int j = pixel[threadId][1];
      int y = imgHeight-1-j;
      cerr << "\nPixel: "
           << i << ", " << j << "       ( -g"
           << i << "," << y << " )"
           << endl;
    }
    
    backTrace();
    feclearexcept(FE_ALL_EXCEPT);
    fedisableexcept(FE_ALL_EXCEPT);
    
    setupSignals();
    signal(SIGFPE, SIG_IGN);

    // kernel bug prevents returning from SIGFPE on linux kernel 2.6.15 (and 2.6.22)
    // http://lkml.org/lkml/2007/8/25/11
    
    static void *last_addr = NULL;
    static int same_count = 0;
    if (last_addr == info->si_addr) {
      if (++same_count >= 5) {
        printf("Exiting after receiving "
               "the same signal %d times\n",
               same_count + 1);
        exit(-1);
      }
    }
    last_addr = info->si_addr;
    return;
  }

  backTrace();

  cerr << "\n" << name << " !\n\n";
  
  if (sig == SIGUSR1){
    for (int i=0; i<NSIG; i++)
      signal(i, SIG_DFL);
    setupSignals();
    return;
  }

  exit(1);
}
	
bool getIsAuthor(){

  struct passwd *pw = getpwuid(geteuid());
  if (!pw)
    return false;
  
  const char *user = pw->pw_name;
  if (!user || !strstr(user, "beason") && !strstr(user, "godvirus"))
    return false;

  return true;
}

