// Title  : pfmtohdr
// Purpose: Converts portable float map (.pfm) file to Radiance format (.hdr)
// Usage  : pfmtohdr <infile.pfm> <outfile.hdr>
// Compile: g++ pfmtohdr.cxx -o obj/pfmtohdr obj/rgbe.o
// Author : Kevin Beason 6/26/07

#include <stdio.h>
#include <errno.h>
#include <iostream>
#ifdef __darwin__
#include <machine/endian.h>
#else
#include <endian.h>
#endif
#include "FileUtils.h"
#include <math.h>

using namespace std;

int main(int argc, char *argv[]){

  const bool verbose = false;
  
  if (argc<3) {
    printf("Usage: pfmtohdr <infile.pfm> <outfile.png>\n",
           argv[0]);
    return 1;
  }

  int dim[2];
  float *depthRGB = FileUtils::readPFM(dim, argv[1], verbose);
  if (!depthRGB){
    printf("Error reading PFM\n");
    return 1;
  }

  // scale depth to be [0,1], disregarding inf values
  int numPix = dim[0]*dim[1];
  float min, max;
  for (int i=0; i<numPix; i++){
    int j = i*3;
    if (isinf(depthRGB[j]))
      continue;
    if (depthRGB[j]<min)
      min=depthRGB[j];
    if (depthRGB[j]>max)
      max=depthRGB[j];
  }
  cerr << "min=" << min << " max=" << max << "\n";
  for (int i=0; i<numPix; i++){
    int j = i*3;
    if (isinf(depthRGB[j])){
      depthRGB[j+0] = 1;
      depthRGB[j+1] = 0;
      depthRGB[j+2] = 0;
    } else {
      float v = (depthRGB[j]-min) / (max-min);
      depthRGB[j+0] = v;
      depthRGB[j+1] = v;
      depthRGB[j+2] = v;
    }
  }
  
  if (verbose) printf("Writing %s...", argv[2]);
  FileUtils::writePNG(dim, argv[2], (const SbColor*)depthRGB, 2.2, "pfmtopng");
  if (verbose) printf("done.\n");
  
  return 0;
}
